// *****************************************************************************
//
//  Routine:      acReadITR
//  Author:       T.Glebe, I.Kisel
//                Modified for the ITR: W. Gradl
//
//  Description:  call the ITR initialization to get the ITR constants
//                - if key not assigned, use KUIP variables as major,minor
//                - returns <0 error
//                           0 no ITR update necessary
//                           1 update done using key
//                           2 update done using date/time or run number
//                
//
//  Modification:	21.03.2000 - added code used by L2FORMARTE (fg sciacca)
//			02.04.2000 - fixed handling of keytable mechanism (fg sciacca)
//			11.04.2000 - bugs removed
//				   - L2FORMARTE code integrated with VDS code (fg sciacca)
//			30.06.2000 - bu fix: now ok with unsparsified data (fg sciacca)
//                      10.08.2000 - WGradl: => ITR
//                      13.11.2000 - TZ: new routine for update of ITR constants using
//                                   keytable mechanism
//                      15.11.2000 - AG: introduced ARTE_mark()
//
// *****************************************************************************

using namespace std;


#include <stdio.h>

#include "arte/arte.hh"
#include "database/keytable.h"
#include "farm/farm.hh"
#include "grr.h"
#include "itrarte/ITR_arte.h"
#include "itrleda.h"
#include "ITR_marvin.h"
#include "itrarte/ITR_steer.h"

// sparse event definition
#include "event.h"
#include "etb.h"

// static variables global for this module
static int ITRserver =	-1;
static int ITRserver_mon = -1;
static int ITRserver_align = -1;
static int ITRtable_chamber =	-1;
static int ITRtable_chip =	-1;
static int ITRtable_nede =	-1;
static int ITRtable_nesl =	-1;
static int ITRtable_fed =	-1;
static int ITRtable_chanstat =	-1;
static int ITRtable_ped =	-1;
static int ITRtable_noise =	-1;
static int ITRtable_align =	-1;


// *****************************************************************
// function
// *****************************************************************

int acReadITR(int KTstatus) 
{
  //==================================================================
  // no action for MC
  //==================================================================
  
  if (ArteTable<MTRA>::size() > 0 ) return 0;

  ARTE_mark( "acReadITR 0", 0 );
    
  //==================================================================
  // database server (default: /ITR_Setup)
  //                  can be changed by setting DB_ITR
  //==================================================================
 
  char default_setup[]   = "/ITR_Setup";
  char default_monitor[] = "/ITR_MONITOR";
  char default_align[]   = "/ITR_Align";

  const char *kSetupEnv = "ITR_SETUPDB";
  const char *kAlignEnv = "ITR_ALIGNDB";
  const char *kMonitorEnv= "ITR_MONITORDB";

  char *dbserver, *dbserver_monitor, *dbserver_align;
  
  static bool firsttime = true;

  if (getenv(kSetupEnv) == NULL){
    dbserver = default_setup;
  } else {
    dbserver = getenv(kSetupEnv);
  }
 
  if (getenv (kAlignEnv) == NULL) {
    dbserver_align = default_align;
  } else {
    dbserver_align = getenv (kAlignEnv);
  }
  
  if (getenv (kMonitorEnv) == NULL) {
    dbserver_monitor = default_monitor;
  } else {
    dbserver_monitor = getenv (kMonitorEnv);
  }

  
  if (firsttime) {
    ITR_leda_param_init();
    ITR_marvin_param_init();
  }

  switch (KTstatus) {

    case 2: 
      // structure of keytable changed, so loop over the keytable and get 
      // your serverid and tableids 

      ITRserver       = KeyFindServer(dbserver); 
      ITRserver_mon   = KeyFindServer(dbserver_monitor); 
      ITRserver_align = KeyFindServer(dbserver_align); 

      if (ITR_DBVERS_MAJ != -1 && ITR_DBVERS_MIN != -1) {
	cout << "acReadITR: === Version of Setup Tables manually set to [" 
	     << ITR_DBVERS_MAJ << ", " << ITR_DBVERS_MIN << "]" << endl;
	cout << "           === Ignoring key table entry" << endl;
	break; // use old method
      }


      if (ITRserver == -1) { 
        cout<<"acReadITR: KeyTable Server " << dbserver << " not found!"<<endl; 
        break;  // use old method
      }
      if (ITRserver_mon == -1) { 
        cout<<"acReadITR: KeyTable Server " << dbserver_monitor << " not found!"<<endl; 
        break;  // use old method
      } 
      if (ITRserver == -1) { 
        cout<<"acReadITR: KeyTable Server " << dbserver_align << " not found!"<<endl; 
        break;  // use old method
      } 

      cout << "acReadITR: Looking for Keytable..." << endl;
      ITRtable_chamber =	KeyFindTable(ITRserver, "ITR/Chamber"); 
      ITRtable_chip =		KeyFindTable(ITRserver, "ITR/Chip"); 
      ITRtable_fed =		KeyFindTable(ITRserver, "ITR/Fed"); 
      ITRtable_nede =		KeyFindTable(ITRserver, "ITR/nede"); 
      ITRtable_nesl =		KeyFindTable(ITRserver, "ITR/nesl"); 
      ITRtable_chanstat =	KeyFindTable(ITRserver_mon, "ITR/Chanstat"); 
      ITRtable_ped =    	KeyFindTable(ITRserver_mon, "ITR/Ped"); 
      ITRtable_noise =  	KeyFindTable(ITRserver_mon, "ITR/Noise"); 
      ITRtable_align =  	KeyFindTable(ITRserver_align, "ITR/Align"); 

      if ((ITRtable_chamber == -1) || (ITRtable_chip == -1) || 
	   (ITRtable_fed == -1) || (ITRtable_chanstat == -1) ||
	   (ITRtable_nede == -1) || (ITRtable_nesl == -1)) { 
        cout<<"acReadITR: Keytable not found!" << endl; 
	break;  // use old method
      }; 
      // NO _BREAK_ HERE! 

    case 1: 
      // an update happened, look if your server is involved 

      if (!KeyServerNew(ITRserver) && !KeyServerNew(ITRserver_align)) {
        cout<<"acReadITR: no ITR update necessary, server not involved"<<endl;      
        cout<<"acReadITR: returning with rc = 0"<<endl;
	ARTE_mark( "acReadITR 1", 0 );
      return 0; //no ITR update necessary
      };

      // now for the tables - look whether they have changed 
      if (KeyTableNew(ITRtable_chip) && ITR_DBVERS_MAJ == -1) { 
        // using LEDA to get the Tables x:
        //cout<<"acReadITR: updating ITR tables"<<endl;
        MizziDatabase(mydb);
        //MizziDatabase(mydb_mon);
        //MizziDatabase(mydb_align);
        Tables x;

	cout<<"acReadITR: keytable found: opening ITR database "<< ITRserver <<endl;
        mydb.open(KeyServerName(ITRserver),DBreadonly);
        x=mydb.get_id();

        ITR_Chamber	->book(mydb);
        ITR_Chip	->book(mydb);
        ITR_Fed		->book(mydb);
        ITR_Chanstat	->book(mydb);
	ITR_nede	->book(mydb);
	ITR_nesl	->book(mydb);

        ITR_Chamber	->remove-> all();
        ITR_Chip	->remove-> all();
        ITR_Fed		->remove-> all();
        ITR_Chanstat	->remove-> all();
	ITR_nede	->remove-> all();
	ITR_nesl	->remove-> all();

        cout << "acReadITR: reading ITR/Chamber [" 
	     << KeyTableMajor (ITRtable_chamber) << ", " 
	     << KeyTableMinor (ITRtable_chamber) << "]" << endl;
        KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_chamber));
        ITR_Chamber	->fetch_data( DBeq );
        ITR_Chamber	->set_keys();

        cout << "acReadITR: reading ITR/Chip [" 
	     << KeyTableMajor (ITRtable_chip) << ", " 
	     << KeyTableMinor (ITRtable_chip) << "]" << endl;
        KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_chip));
        ITR_Chip	->fetch_data( DBeq );
        ITR_Chip	->set_keys();

        cout << "acReadITR: reading ITR/Fed [" 
	     << KeyTableMajor (ITRtable_fed) << ", " 
	     << KeyTableMinor (ITRtable_fed) << "]" << endl;
        KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_fed));
        ITR_Fed		->fetch_data( DBeq );
        ITR_Fed		->set_keys();

        cout << "acReadITR: reading ITR/nede [" 
	     << KeyTableMajor (ITRtable_nede) << ", " 
	     << KeyTableMinor (ITRtable_nede) << "]" << endl;
        KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_nede));
        ITR_nede	->fetch_data( DBeq );
        ITR_nede	->set_keys();

        cout << "acReadITR: reading ITR/nesl [" 
	     << KeyTableMajor (ITRtable_nesl) << ", " 
	     << KeyTableMinor (ITRtable_nesl) << "]" << endl;
        KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_nesl));
        ITR_nesl	->fetch_data( DBeq );
        ITR_nesl	->set_keys();

        cout << "acReadITR: reading ITR/Chanstat [" 
	     << KeyTableMajor (ITRtable_chanstat) << ", " 
	     << KeyTableMinor (ITRtable_chanstat) << "]" << endl;
        KeyTableStamp(x,KeyServerName(ITRserver_mon),KeyTableName(ITRtable_chanstat));
        ITR_Chanstat	->fetch_data( DBeq );
        ITR_Chanstat	->set_keys();

	ITR_Chamber	->fetch_relations(DBeq); /* does not load any table */
	ITR_Fed		->fetch_relations(DBeq); /* does not load any table */
	ITR_Chanstat	->fetch_relations(DBeq);
	ITR_nede	->fetch_relations(DBeq);
	ITR_nesl	->fetch_relations(DBeq);
	ITR_Chip	->fetch_relations(DBeq); /* loads fed and side relations */

        ITR_Chamber	->unbook();
        ITR_Chip	->unbook();
        ITR_Fed		->unbook();
        ITR_nede	->unbook();
        ITR_nesl	->unbook();
        ITR_Chanstat	->unbook();


        if(firsttime) {	
	  firsttime = false;
	  ITR_leda_startprogram (-1, -1);
        }

	ITR_marvin_init();
	cout<<"acReadITR: ITR tables loaded: closing ITR database "<< ITRserver <<endl;
      }  

      if (KeyTableNew(ITRtable_align) && ITR_ALIGN_MAJ == -1) {
	int alignMajor, alignMinor;
	alignMajor = KeyTableMajor (ITRtable_align);
	alignMinor = KeyTableMinor (ITRtable_align);
	cout << "acReadITR: reading ITR/Align information from Server " <<
	  dbserver_align << " for [" << alignMajor << "," << alignMinor << "]" << endl;
	ITR_leda_fetch_alignment (dbserver_align, alignMajor, alignMinor); 
      }
      else if (ITR_ALIGN_MAJ != -1) {
	cout << "acReadITR: === Version of alignment table set manually to [" 
	     << ITR_ALIGN_MAJ << "," << ITR_ALIGN_MIN << "]" << endl;
	cout << "           === Ignoring key table entry " << endl;
	
	ITR_leda_fetch_alignment (dbserver_align, ITR_ALIGN_MAJ, ITR_ALIGN_MIN);
      }

      ITR_arte_updateGEDE();

      cout<<"acReadITR: returning with rc = 1"<<endl;
      ARTE_mark( "acReadITR 1", 0 );
      return 1; // update using keytable


      //break; 
  } // of switch 

  // No keytable information available, using old method...

  if(firsttime) {
    cout << "acReadITR: Key NOT assigned --> using old method" << endl; 
  }
  
  static int ITR_Chamber_major = -1;
  static int ITR_Chamber_minor = -1;
  static bool findmajmin = true;
  static MizziDatabase( itrdb_setup );

  // get event pointer
  EVENT_HDR *event = ARTE_get_event();
  if ( !event ) {
    cout << "acReadITR: Could not find event!" << endl;
    ARTE_mark( "acReadITR 1", 0 );
    return -1; // error
  }      
  
  //    Get constants from database
  //
  // Time stamp is evaluated in this order:
  //  1. date/time   from itrleda IOL
  //  2. major/minor from itrleda IOL
  //  3. event time from DDAQ (if all IOL parameters are =0)
  int major = 0;
  int minor = 0;
  int ndaq;
  
//   EVT_DATA_RECORD daq;
//   ETB_get_daqs( event, 1, &daq, &ndaq );
//   if ( ndaq > 0 ) {
//     major = daq.data.event_time;
//     minor = 0;
//   }

  if (ITR_DBVERS_MAJ != -1 && ITR_DBVERS_MIN != -1) {
    findmajmin = false;
    major = ITR_DBVERS_MAJ;
    minor = ITR_DBVERS_MIN;
  } else {
    int dbmajor, dbminor;

    if (ITR_STEERDB_MAJ != -999 && ITR_STEERDB_MIN != -999) {
      dbmajor = ITR_STEERDB_MAJ;
      dbminor = ITR_STEERDB_MIN;   
    } else {
      if( ArteTable<GEHD>::size() ){
	ArtePointer<GEHD> gehd( 1 );  
	if (int(gehd->gevs[2]) >  0) {ITR_STEERDB_MAJ=2; ITR_STEERDB_MIN=2;}
	else if (int(gehd->gevs[2]) == 0) {ITR_STEERDB_MAJ=2000; ITR_STEERDB_MIN=2;}
      }
      dbmajor = ITR_STEERDB_MAJ;
      dbminor = ITR_STEERDB_MIN;
    }

    cout << "acReadITR:  *** using default setup as found in /ITR_Setup:ITR/Steering ["
	 << ITR_STEERDB_MAJ << ", " << ITR_STEERDB_MIN << "]" 
	 << endl;
    
    iol_scan_preferences_from_db ("/ITR_Setup:ITR/Steering", dbmajor,
				  dbminor);
    major = 0; 
    minor = 0;
    findmajmin = false;
  } 


  if(ITR_Chamber_major == -1 || 
     (major > ITR_Chamber_major && !findmajmin)) {

    // setup iol parameters for (old style) database access
    extern Itrleda_Iol_t itrleda_iol;

    /* decide which alignment version to take: 
     *  1. explicitly set with KUIP commands? 
     *  2. implicitly by setting setup major/minor? 
     *  3. no versions set -> take those of steering dbase read
     */
    if (ITR_ALIGN_MAJ != -1) {               // 1.
	cout << "acReadITR:  *** Version of alignment table set manually to [" 
	     << ITR_ALIGN_MAJ << "," << ITR_ALIGN_MIN << "]" << endl;

	itrleda_iol.align_major.value = ITR_ALIGN_MAJ;
	itrleda_iol.align_minor.value = ITR_ALIGN_MIN;
    } else if (major != 0 && minor != 0) {   // 2.
 	itrleda_iol.align_major.value = major;
	itrleda_iol.align_minor.value = minor;
    } else {                                 // 3.
      /* do nothing */
    }
    


    int rc;
    
    rc = ITR_leda_startprogram (major, minor);

    if (rc != GRR_OK) {
      ITR_HITPREP_FAILED=1;
      ARTE_mark( "acReadITR 1", 0 );
      return (-1);
    }

    ITR_marvin_init();

    ITR_arte_updateGEDE();

    if(firsttime) {		//silraw and marvin init must occurr after DB access

      firsttime = false;
    }

  } else {
    ARTE_mark( "acReadITR 1", 0 );
    return 0; // no update necessary
  }

  // find major minor until which dataset is valid
  if(ITR_Chamber_major == -1  || 
     findmajmin            ||
     major > ITR_Chamber_major ) {

    //    itrdb_setup.open( itrleda_iol.setup_database.value, 0 );
    itrdb_setup.open( dbserver, 0 );
    itrdb_setup.stamp( major, minor );
    ITR_Chamber    ->book( itrdb_setup );
    if( ITR_Chamber->match( DBgt ) == 1 ) {
      // successfull match
      ITR_Chamber_major = TableMajor( itrdb_setup.get_id() );
      ITR_Chamber_minor = TableMinor( itrdb_setup.get_id() );
      findmajmin = false;
    } else {
      //if there is no newer table, lookup would take place in each event!
      //      findmajmin = true;
      ITR_Chamber_major = major + 200;
      findmajmin = false;
    }

    ITR_Chamber->unbook();
    itrdb_setup.close();
  }

  cout<<"acReadITR: returning with rc = 2"<<endl;
  ARTE_mark( "acReadITR 1", 0 );
  return 2;  // constants read in using old method 
}
