
// *********************************************************************
//
//  Routine:      acread_
//  Author:       T.Oest
//
//  Description:  general steering routine to apply internal and
//                
//
//  Modification:
//  - acGlobAlChange() switched off (causes a segmentation fault
//    if the database is not defined)                   T.Oest  2/8/99
//  - VDS and global alignment constants read in        T.Oest 16/9/99
//  - keytable code by M. Hissmann added                T.Oest 20/9/99
//  - call to acReadTARGET added                        T.Oest 24/9/99
//  - call of acReadDQ added                            T.Oest 29/10/99
//  - alignment is read in only for DST if reco on or
//    KeyBook was called in the kumac                   T.Oest 02/12/99
//  - set alca flag, add routine to return alcaFlag     
//    call acReadMuon and ITR_arte_updateGEDE           T.Oest 10/02/00
//  - acSetDateTime() included, GTAR update moved after global
//    acReadGlobal                                      T.Oest 08/03/00 
//  - of only HITB preparation is requested: read in 
//    only those constants which are used               T.Oest 27/03/00 
//  - acReadECAL included                               T.Oest 31/03/00
//  - acReadTARGET called for every event               T.Oest 11/04/00
//  - use ARTE_mark() to save position                  AG     14/04/00
//  - more ARTE_mark()                                  AG     25/04/00
//  - use faComp() to check components                  AG     11/05/00
//  - more ARTE_mark()                                  AG      2/06/00
//  - polishing                                         AG      6/06/00
//  - call of new acReadITR routine added               TZ     13/10/00
//
// ********************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "database/keytable.h"
#include "component.h"
#include "reco/rcflags.hh"
#include "itrarte/ITR_arte.h"

// external flags
extern int KeyBookWasCalled;

extern int acFlagReadVDS ;
extern int acFlagReadOTR ;
extern int acFlagReadITR ;
extern int acFlagReadECAL;
extern int acFlagReadMUON;
extern int acFlagReadRICH;

// prototypes
int acReadTARGET(int);
int acReadVDS(int);
int acReadITR(int);
int acReadOTR(int);
int acReadRICH(int);
int acReadECAL(int);
int acReadGlobal(int);
int acReadDQ(int);

extern "C" int acReadMuon(int);

void acGlobalAlignmentForGEDE(unsigned int);
void acSetDateTime();

// globals
static          int firstCall         = 0;
static          int readAlignment     = 0;
static unsigned int alcaFlag          = 0;

//*****************************************************************
// function
//*****************************************************************

extern "C" void acread_() {

  // set mark
  ARTE_mark( "acread_ 0", 0 );

//==================================================================
// no action for MC
//==================================================================

  if (ArteTable<MTRA>::size() > 0 ) return ;

//==================================================================
// 1. set date and time in event header
//==================================================================

  acSetDateTime();

//==================================================================
// check for the first call if alignment should be read in
//   - no reading  for MINIs (ArteTable<DEVT>::size() == 0)
//   - reading     if reconstruction is on or KeyBookWasCalled == 1
//   - only used subdetectors if nothing than hitb is filled 
//==================================================================

  if (firstCall == 0){

   firstCall = 1;

   if ((recoFlag   != 0  || KeyBookWasCalled == 1) &&
        ArteTable<DEVT>::size() != 0){
    if (recoHITB != 0 && recoPid == 0){
         readAlignment = 1;
    } else {
         readAlignment = 2;
    };
   };

  }

//==================================================================
// 1. check for a new alignment/calibration key (DKEY table)
//      KTstatus will contain status of KeyTable
//               -2  no KUIP keytable initalisation
//               -1  failed to find keytable 
//               >0  new keytable
//==================================================================

  int KTstatus = -2;
  if (KeyBookWasCalled == 1) KTstatus = KeyLoadTables(faEVHDdbmajor());


  if (KTstatus==-1) {
      // just protest loudly ! 
      // the keytable with version could not be found
      // faEVHDdbmajor() faEVHDdbminor()
      KTstatus=0;
     }


//==================================================================
// 2. ARTE steering
//==================================================================

  int flagDQ = 0;
  if (KTstatus > 0){
    ARTE_mark( "acReadDQ", 0 );
    flagDQ = acReadDQ(KTstatus);
  }

//==================================================================
// 3. set alignment constants, transforms NEDE to GEDE tables
//  - no action if no reconstruction will be done
//  - alignment sensitive parameters should be taken from 
//    the subdetector database (no off values with respect to NEDE)
//  - constant values (pitch) should be taken from NEDE
//  - return flag: > 0   constants were changed
//                 = 0   no change of constants
//  - return code  < 0   error, constants not found or read error
//                 = 0     constants uptodate
//                 = 1     constants read in by key mechanism
//                 = 2     constants read in using event time or run number
//==================================================================

  int flagTARGET = 0;
  int flagVDS    = 0;
  int flagITR    = 0;
  int flagOTR    = 0;
  int flagMUON   = 0;
  //int flagTRD    = 0;
  int flagRICH   = 0;
  int flagECAL   = 0;
  //int flagHPT    = 0;
  int flagGLOBAL = 0;

  if (KTstatus!=0) {
 
    if (readAlignment != 0)  {    

      // VDS
      int readVDS = faComp( COMP_M_SVD ) && (readAlignment == 2 || recoVDS != 0) && acFlagReadVDS == 1;
      if (readVDS){
	ARTE_mark( "acReadVDS", 0 );
	flagVDS = acReadVDS(KTstatus);
      }

      // ITR
      int readITR = faComp( COMP_M_ITR ) && (readAlignment == 2 || recoITR != 0) && acFlagReadITR == 1;
      if (readITR){
	ARTE_mark( "acReadITR", 0 );
	flagITR = acReadITR(KTstatus);
      }

      // OTR
      int readOTR = faComp( COMP_M_OTR ) && (readAlignment == 2 || recoOTR != 0) && acFlagReadOTR == 1;
      if (readOTR){
	ARTE_mark( "acReadOTR", 0 );
	flagOTR = acReadOTR(KTstatus);
      }

      // RICH
      int readRICH = faComp( COMP_M_RICH ) && (readAlignment == 2 || recoRICH != 0 || recoPidRICH != 0) &&
	acFlagReadRICH == 1;      if (readRICH){
	ARTE_mark( "acReadRICH", 0 );
	flagRICH = acReadRICH(KTstatus);
      }

      // ECAL
      int readECAL = faComp( COMP_M_ECAL ) && (readAlignment == 2 || recoECAL != 0 || recoPidECAL != 0) &&
	acFlagReadECAL == 1;
      if (readECAL){
	ARTE_mark( "acReadECAL", 0 );
	flagECAL = acReadECAL(KTstatus);
      }

      // MUON
      int readMUON = faComp( COMP_M_MUON ) && (readAlignment == 2 || recoMUON != 0 || recoPidMUON != 0) &&
	acFlagReadMUON == 1;
      if (readMUON){
	ARTE_mark( "acReadMuon", 0 );
	flagMUON = acReadMuon(KTstatus);
      }

      // GLOBAL
      ARTE_mark( "acReadGlobal", 0 );
      flagGLOBAL = acReadGlobal(KTstatus);
    }

  }

  // TARGET
  ARTE_mark( "acReadTARGET", 0 );
  flagTARGET = acReadTARGET(KTstatus);
  
//==================================================================
// 4. set alca flag
//    subdetectors are identified by the DAQ indentifier (component.h)
//          COMP_M_TARG
//          COMP_M_SVD
//          COMP_M_ITR
//          COMP_M_OTR
//          COMP_M_HIPT
//          COMP_M_MUON
//          COMP_M_RICH
//          COMP_M_ECAL
//          COMP_M_TRD
//          0x00000001  -> global alignment
//          
//     the following bits are set in alcaFlag
//         COMP_M_xxx        if constants are avaiable/read in
//         COMP_M_xxx << 16  if the key mechanism has been used
//     
//==================================================================

#define COMP_M_GLOBAL   0x00000001

  unsigned int alcaUpdate = 0;

  if (flagTARGET >= 1) alcaUpdate |=  COMP_M_TARG;
  if (flagVDS    >= 1) alcaUpdate |=  COMP_M_SVD;
  if (flagITR    >= 1) alcaUpdate |=  COMP_M_ITR;
  if (flagOTR    >= 1) alcaUpdate |=  COMP_M_OTR;
  if (flagRICH   >= 1) alcaUpdate |=  COMP_M_RICH;
  if (flagECAL   >= 1) alcaUpdate |=  COMP_M_ECAL;
  if (flagMUON   >= 1) alcaUpdate |=  COMP_M_MUON;
  if (flagGLOBAL >= 1) alcaUpdate |=  COMP_M_GLOBAL;

                       alcaFlag   |= alcaUpdate;

  if (flagTARGET <  0) alcaFlag   &= ~COMP_M_TARG;
  if (flagVDS    <  0) alcaFlag   &= ~COMP_M_SVD;
  if (flagITR    <  0) alcaFlag   &= ~COMP_M_ITR;
  if (flagOTR    <  0) alcaFlag   &= ~COMP_M_OTR;
  if (flagRICH   <  0) alcaFlag   &= ~COMP_M_RICH;
  if (flagECAL   <  0) alcaFlag   &= ~COMP_M_ECAL;
  if (flagMUON   <  0) alcaFlag   &= ~COMP_M_MUON;
  if (flagGLOBAL <  0) alcaFlag   &= ~COMP_M_GLOBAL;

  if (flagTARGET == 1) alcaFlag   |=   COMP_M_TARG   << 16;
  if (flagVDS    == 1) alcaFlag   |=   COMP_M_SVD    << 16;
  if (flagITR    == 1) alcaFlag   |=   COMP_M_ITR    << 16;
  if (flagOTR    == 1) alcaFlag   |=   COMP_M_OTR    << 16;
  if (flagRICH   == 1) alcaFlag   |=   COMP_M_RICH   << 16;
  if (flagECAL   == 1) alcaFlag   |=   COMP_M_ECAL   << 16;
  if (flagMUON   == 1) alcaFlag   |=   COMP_M_MUON   << 16;
  if (flagGLOBAL == 1) alcaFlag   |=   COMP_M_GLOBAL << 16;

  if (flagTARGET == 2) alcaFlag   &= ~(COMP_M_TARG   << 16);
  if (flagVDS    == 2) alcaFlag   &= ~(COMP_M_SVD    << 16);
  if (flagITR    == 2) alcaFlag   &= ~(COMP_M_ITR    << 16);
  if (flagOTR    == 2) alcaFlag   &= ~(COMP_M_OTR    << 16);
  if (flagRICH   == 2) alcaFlag   &= ~(COMP_M_RICH   << 16);
  if (flagECAL   == 2) alcaFlag   &= ~(COMP_M_ECAL   << 16);
  if (flagMUON   == 2) alcaFlag   &= ~(COMP_M_MUON   << 16);
  if (flagGLOBAL == 2) alcaFlag   &= ~(COMP_M_GLOBAL << 16);


//==================================================================
// update GEDE tables with global constants
//==================================================================

  if (alcaUpdate != 0){
    ARTE_mark( "acGlobalAlignmentForGEDE", 0 );
    acGlobalAlignmentForGEDE(alcaUpdate);
  }

//==================================================================
// 
//==================================================================
  ARTE_mark( "acread_ 1", 0 );
  return;
}
//*****************************************************************
// function to return alcaFlag
//*****************************************************************

unsigned int acFlag() {

 return alcaFlag;

};


