/***********************************************************************/
/*  isaelectron                                                        */
/*                                                                     */
/*  intention: electron selection using RICH                           */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate                            */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is not designed for electron identification         */
/*            so this is a background removal routine                  */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates:                                                           */
/*                                                                     */
/***********************************************************************/

#include "arte/arte.hh"
#include <iostream.h>

int anl_isaelectron(ArtePointer<RTRA> candidate) {

  float electron_thres = 0.;  // electron threshold
  float back_cut = 0.5;       // background likelyhood cut

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else
    momentum = fabs(1./candidate->pf);

  if(momentum > electron_thres) { // electron threshold at very low momentum
    if (rtri->flt > 0) { // RISE considers the track inside RICH acceptance
      accept = 0;
      if (candidate->crich > 0)  // RISE found a ring for this track
	accept = 1;
    }

    if (accept != 1) { // RISE is not 100% efficient to find rings. Let's try with RITER.
      if (rtri->fls > 0) { // RITER considers the track inside RICH acceptance
	accept = 0;
	float like = rtri->lspi + rtri->lsmu + rtri->lse;
	if (like > back_cut)  // RITER sees a signal
	  accept = 1;
      }
    }
  }

  return(accept);
}



/***********************************************************************/
/*  isakaon                                                            */
/*                                                                     */
/*  intention: kaon selection using RICH                               */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate                            */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is designed for kaon/pion separation                */
/*            so this is a selction routine based on particle Id       */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates:                                                           */
/*                                                                     */
/***********************************************************************/

int anl_isakaon(ArtePointer<RTRA> candidate) {

  float pion_thres = 3.;   // pion threshold
  float kaon_thres = 10.;  // kaon threshold
  float lowp_cut = 0.9;    // cut on likelyhood for low momentum
  float highp_cut = 0.5;   // cut on likelyhood for high momentum

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else 
    momentum = fabs(1./candidate->pf);

  if (rtri->fls > 0) { // RITER considers the track inside RICH acceptance
    float cand_like = rtri->lsk;
    float normalization = cand_like + rtri->lspi;
    if(normalization == 0.) normalization = 1.;
    float like = cand_like/normalization;


    if (momentum > pion_thres) {
      accept = 0;
      if ((momentum < kaon_thres && like > lowp_cut) || (momentum >= kaon_thres && like > highp_cut))
	accept = 1;
    }
  }

  return(accept);
}




/***********************************************************************/
/*  isamuon                                                            */
/*                                                                     */
/*  intention: muon selection using RICH                               */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate                            */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is not designed for muon identification             */
/*            so this is a background removal routine                  */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates:                                                           */
/*                                                                     */
/***********************************************************************/

int anl_isamuon(ArtePointer<RTRA> candidate) {

  float muon_thres = 3.; // muon threshold
  float back_cut = 0.5;  // background likelyhood cut

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else 
    momentum = fabs(1./candidate->pf);

  if(momentum > muon_thres) { // muon threshold at 3 GeV
    if (rtri->flt > 0) { // RISE considers the track inside RICH acceptance
      accept = 0;
      if (candidate->crich > 0)  // RISE found a ring for this track
	accept = 1;
    }

    if (accept != 1) { // RISE is not 100% efficient to find rings. Let's try with RITER as well.
      if (rtri->fls > 0) { // RITER considers the track inside RICH acceptance
	accept = 0;
	float like = rtri->lspi + rtri->lsmu + rtri->lse;
	if (like > back_cut)  // RITER sees a signal
	  accept = 1;
      }
    }
  }

  return(accept);
}



/***********************************************************************/
/*  isapion                                                            */
/*                                                                     */
/*  intention: pion selection using RICH                               */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate                            */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is not designed for pion identification             */
/*            so this is a background removal routine                  */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates:                                                           */
/*                                                                     */
/***********************************************************************/

int anl_isapion(ArtePointer<RTRA> candidate) {

  float pion_thres = 3.;  // pion threshold
  float back_cut = 0.5;   // background likelyhood cut

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else 
    momentum = fabs(1./candidate->pf);

  if(momentum > pion_thres) { // pion threshold at 3 GeV
    if (rtri->flt > 0) { // RISE considers the track inside RICH acceptance
      accept = 0;
      if (candidate->crich > 0)  // RISE found a ring for this track
	accept = 1;
    }

    if (accept != 1) { // RISE is not 100% efficient to find rings. Let's try with RITER as well.
      if (rtri->fls > 0) { // RITER considers the track inside RICH acceptance
	accept = 0;
	float like = rtri->lspi + rtri->lsmu + rtri->lse;
	if (like > back_cut)  // RITER sees a signal
	  accept = 1;
      }
    }
  }

  return(accept);
}



/***********************************************************************/
/*  isaproton                                                          */
/*                                                                     */
/*  intention: proton selection using RICH                             */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate                            */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is good for proton/pion separation                  */
/*            so this is a selection routine based on particle Id      */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates:                                                           */
/*                                                                     */
/***********************************************************************/

int anl_isaproton(ArtePointer<RTRA> candidate) {

  float pion_thres = 3.;     // pion threshold
  float proton_thres = 20.;  // proton threshold
  float lowp_cut = 0.9;      // cut on likelyhood for low momentum
  float highp_cut = 0.5;     // cut on likelyhood for high momentum

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else 
    momentum = fabs(1./candidate->pf);

  if (rtri->fls > 0) { // RITER considers the track inside RICH acceptance
    float cand_like = rtri->lsp;
    float normalization = cand_like + rtri->lspi;
    if(normalization == 0.) normalization = 1.;
    float like = cand_like/normalization;


    if (momentum > pion_thres) {
      accept = 0;
      if ((momentum < proton_thres && like > lowp_cut) || (momentum >= proton_thres && like > highp_cut))
	accept = 1;
    }
  }

  return(accept);
}








