#include "arte/arte.hh"
#include "grover/grover.hh" // - class defs, f.prototypes and global variables

// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 00-11-01  M.Sang   No RTRA corresponding to primary (makes no sense!)
// 00-11-08  M.Sang   Writes chi2 of vertex into RVER; bugfix in cov matrix
// 00-11-13  M.Sang   Now writes vtx 'parent' into RTRA if secondary.
// 00-11-13  M.Sang   Bug fix in momentum writing
// 00-11-15  M.Sang   Bug fix in safety check

int GroverAddToRVER(GroverVertexList* vertices,
		    int type = 1) 
{ //
  // Type = 1 for primaries, 2 for secondaries
  // M.Sang 05/06/00
  //

  if (vertices == 0) return -1;
  if (!(type == 1 || type == 2)) return -2;

  bool verbose = false;//(groverCuts.Verbose == 1);
  RVER vertex;
  RTRA track;
  int charge = 0;

  for (GroverVertexList::iterator ivert = vertices->begin();
       ivert != vertices->end(); ++ivert) {
    
    // first change the coordinate system
    GroverVertex tempvtx = *ivert;
    tempvtx.LocalToGlobal();


    // for secondaries, check they've been fit !
    bool skip = false;
    if (type == 2) {
      if (tempvtx.Momentum() == -999.0) {
	if (tempvtx.Size()>1 && tempvtx.Size()<11) {
	  tempvtx.Fit(-1.0); 
	  cout << "*** GroverAddToRVER: WARNING - refitting secondary "
	          "vertex with current constraints!" << endl;
	}
	else {
	  cout <<"*** GroverAddToRVER: WARNING - "
	         "strange unfitted secondary passed - ignored" << endl;
	  skip = true;
	}
      }
    }
    if (skip) continue;


    // Simple RVER stuff first
    vertex.xv=tempvtx.X();   vertex.yv=tempvtx.Y();   vertex.zv=tempvtx.Z();
    for (int i=0;i<6;i++) vertex.cv[i] = tempvtx.Cov(i);
    vertex.nt   = tempvtx.particles.size();
    vertex.mass = tempvtx.Mass();
    vertex.merr = tempvtx.MassError();
    vertex.wire = tempvtx.Wire();
    vertex.chi2 = tempvtx.Chi2();



    // RTRA stuff
    ArtePointer<RTRA> my_btra = 0;
    if (type == 2) {
      float mom = tempvtx.Momentum();
      if (mom == 0.0) track.pf = 0.0;
      else {
	track.pf = 1.0/mom;
	if (charge<0) track.pf = -1.0*track.pf;
      }
      GroverParticle ttrk(&tempvtx);
      track.xf = ttrk.BeginX();  track.yf = ttrk.BeginY();  track.zf = ttrk.BeginZ();
      track.txf = ttrk.BeginSlopeX();  track.tyf = ttrk.BeginSlopeY();
      // off-diagonals
      for (int i=0; i<14; i++) track.cvf[i] = ttrk.BeginCov(i);
      track.cvf[14] = ((mom*mom)>0.0) ? ttrk.BeginCov(14)/(mom*mom) : 1000000.0;
      track.chi = tempvtx.Chi2();
      // end vertex pointer
      track.ever= ++(ArtePointer<RVER>(ArteTable<RVER>::size()));
      track.fit = 3;// Flagged as a long-lived decay
      if (track.chi==0.0) track.fit += -100;
      track.fle = track.flm = track.flr = track.flt 
	= track.hca = track.hin = track.hmu = track.hou 
	= track.hri = track.hsi = track.htr = 0;
      track.lee = track.leh = track.lemi = track.lmh 
	= track.lmmu = track.lre = track.lrk = track.lrmu 
	= track.lrp = track.lrpi = track.lte = track.lth = 0.0;
      track.topp = 0;
      //track.topp = pointer to TOPP row for a J/Psi etc. used for the 
      //constraint -> if not constrained fit , set to zero.
      ArteTable<RTRA>::push_back(track);
      my_btra = ArtePointer<RTRA>(ArteTable<RTRA>::size());
    }


    vertex.btra = my_btra;
    ArteTable<RVER>::push_back(vertex);


    // Next, fill the relations between RVER and RTRAs (if possible)
    ArtePointer<RVER> pvertex(ArteTable<RVER>::size()); 
    for (GroverParticleList::iterator ipar = ivert->particles.begin();
	 ipar != ivert->particles.end(); ++ipar) {
      charge += ipar->Charge();
      if (ipar->Type() == 1) {
	// This GroverParticle was made from an entry in RTRA. 
	// Good! we can fill the relation
	ArtePointer<RTRA> parent = ipar->ParentRTRA();
	if (parent != 0) {
	  if (verbose) cout << "*** GroverAddToRVER - setting relation" << endl;
	  RREL relation;
	  relation.chi  = ipar->Chi2();
	  relation.flag = 1;
	  ArteTable<RREL>::push_back(relation);
	  ArtePointer<RREL> prelation(ArteTable<RREL>::size());
	  createRelation(pvertex, prelation, &RVER::rrel);
	  createRelation(parent,  prelation, &RTRA::rrel);
	}
	else {
	  if (verbose) 
	    cout << "*** GroverAddToRVER - WARNING: RTRA parent of a vertex track "
		 << " was not set!" << endl;
	}
      }
    }


  } // finished loop over vertex list

  return 0;
}
