#include "arte/arte.hh"
#include "grover/grover.hh"
#include <math.h>
#include "TROOT.h"        
#include "TApplication.h" 
#include "TCanvas.h"      
#include "TCONE.h"        
#include "TTUBE.h"        
#include "TBRIK.h"        
#include "TNode.h"        
#include "TRotMatrix.h"   


extern int xargc;    // # of command line arguments of main()
extern char** xargv; // list of -"-

extern int ConvertAngles(float tx, float ty, 
			 double& theta1, double& phi1,
			 double& theta2, double& phi2,
			 double& theta3, double& phi3);

extern void rnorml(float&);

int GroverDraw3D(GroverParticleList* Particles,
		 GroverWireList*     Wires,
		 GroverVertexList*   Primaries,
		 GroverVertexList*   Secondaries,
		 GroverVertexList*   MCPrimaries = 0,
		 GroverVertexList*   MCSecondaries = 0,
		 GroverParticleList* Specials = 0,
		 GroverCuts*         usercuts = 0)
{
  ///////////////////////////////////////////////////////////////
  //
  // GroverDraw3D
  // M.Sang 11 Oct 00
  //
  // Simple 3D event display function for Grover.
  // Uses ROOT TGeometry classes and X3D interface. 
  // See http://root.cern.ch for more information.
  //
  // Modifications:
  //
  // 00-10-11 Max    First version - works OK, track drawing
  //                 disabled (still buggy and I'm away next week)
  // 00-11-01 Max    Works fine now
  // 00-11-08 Max    Added facility to only draw tracks with certain 
  //                 minimum p, pt, Nhits, or maximum chi2/dof
  // 00-11-14 Max    chi2 cut bug fix
  //
  ///////////////////////////////////////////////////////////////


  // 1) Cuts
  static GroverCuts localcuts;
  GroverCuts* cuts = &localcuts;
  if (usercuts != 0) cuts = usercuts;
  bool verbose          = (cuts->Verbose == 1);
  bool drawprim         = (cuts->TDPrimaries == 1);
  bool drawsec          = (cuts->TDSecondaries == 1);
  bool drawwires        = (cuts->TDWires == 1);
  bool drawaxes         = (cuts->TDAxes == 1);
  bool drawtr           = (cuts->TDTargetRegion == 1);
  bool drawmcprim       = (cuts->TDMCPrimaries == 1);
  bool drawmcsec        = (cuts->TDMCSecondaries == 1);
  bool drawprimtracks   = (cuts->TDPrimaryTracks == 1);
  bool drawsectracks    = (cuts->TDSecondaryTracks == 1);
  bool drawunusedtracks = (cuts->TDUnusedTracks == 1);
  bool drawspecialtracks= (cuts->TDSpecialTracks == 1);
  if ( !(drawprim         || drawsec       || drawwires        || 
	 drawaxes         || drawmcprim    || drawmcsec        || 
	 drawprimtracks   || drawsectracks || drawunusedtracks || 
	 drawspecialtracks                                       )
      ) return 0;





  // 2) Constants
  const float nsig                 =   3.5; // number of sigma for box size
  const float SCALE                = 500.0; // for X3D step size
  const float PRIMARYTRACKLENGTH   = (cuts->TDPrimaryTrackLength);
  const float SECONDARYTRACKLENGTH = (cuts->TDSecondaryTrackLength);
  const float UNUSEDTRACKLENGTH    = (cuts->TDUnusedTrackLength);
  const float SPECIALTRACKLENGTH   = (cuts->TDSpecialTrackLength);
  const float XAXISLENGTH          = (cuts->TDXAxisLength);
  const float YAXISLENGTH          = (cuts->TDYAxisLength);
  const float ZAXISLENGTH          = (cuts->TDZAxisLength);
  const float AXISRADIUS           = (cuts->TDAxisRadius);
  const int   TARGETCOLOUR         = (cuts->TDWireColour);
  const int   PRIMARYCOLOUR        = (cuts->TDPrimaryColour);
  const int   SECONDARYCOLOUR      = (cuts->TDSecondaryColour);
  const int   MCPRIMARYCOLOUR      = (cuts->TDMCPrimaryColour);
  const int   MCSECONDARYCOLOUR    = (cuts->TDMCSecondaryColour);
  const int   PRIMARYTRACKCOLOUR   = (cuts->TDPrimaryTrackColour);
  const int   SECONDARYTRACKCOLOUR = (cuts->TDSecondaryTrackColour);
  const int   UNUSEDTRACKCOLOUR    = (cuts->TDUnusedTrackColour);
  const int   SPECIALTRACKCOLOUR   = (cuts->TDSpecialTrackColour);
  const int   TRACKSMOOTHNESS      = (cuts->TDTrackSmoothness);
  const int   MINNH                = (cuts->TDTrackMinNh);
  const float PCUT                 = (cuts->TDTrackMinP);
  const float PTCUT                = (cuts->TDTrackMinPt);
  const float CHI2CUT              = (cuts->TDTrackMaxChi2);



  // 3) Initialisation
  static bool firstcall = true;
  if (firstcall || verbose) 
    cout << "*** GroverDraw3D: Setting up ROOT/X3D display..." << endl;
  if (firstcall || verbose) 
    cout <<"  1) Grover ROOT object..."<<endl;
  static TROOT GroverRoot("GroverRoot", "Grover ROOT object");
  if (firstcall || verbose)
    cout <<"  2) Grover TApplication object..."<<endl;
  static TApplication* GroverTApp = 
    new TApplication("GroverTApp",&xargc,xargv);
  if (firstcall || verbose) 
    cout <<"  3) Grover TCanvas..."<<endl;
  static TCanvas* GroverCanvas = 
    new TCanvas("GroverCanvas","Grover Vertex Display",200,10,700,500);
  if (firstcall||verbose) cout <<"     ...done!"<<endl;




  // 4) Define mother volume
  TBRIK Grover_tr("","","void",
		  XAXISLENGTH*SCALE/2.0,YAXISLENGTH*SCALE/2.0,ZAXISLENGTH*SCALE/2.0);
  Grover_tr.SetVisibility(((drawtr) ? 1 : 0));
  TNode MotherVolume("Mother","Mother volume",&Grover_tr);
  MotherVolume.cd(); // all other volumes contained in here now




  // 5)  Define volumes for target region and axis arrows
  TTUBE Grover_xaxis("","","void",
		     0,AXISRADIUS*SCALE,XAXISLENGTH*SCALE/2.0);
  TCONE Grover_xmrkr("", "","void",
		     0.2*SCALE,0.0,0.2*SCALE,0.0,0.0);
  TTUBE Grover_yaxis("","","void",
		     0,AXISRADIUS*SCALE,YAXISLENGTH*SCALE/2.0);
  TCONE Grover_ymrkr("", "","void",
		     0.2*SCALE,0.0,0.2*SCALE,0.0,0.0);
  TTUBE Grover_zaxis("","","void",
		     0,AXISRADIUS*SCALE,ZAXISLENGTH*SCALE/2.0);
  TCONE Grover_zmrkr("", "","void",
		     0.2*SCALE,0.0,0.2*SCALE,0.0,0.0);
  // cosmetic
  Grover_xaxis.SetNumberOfDivisions(4);
  Grover_yaxis.SetNumberOfDivisions(4);
  Grover_zaxis.SetNumberOfDivisions(4);
  Grover_xaxis.SetLineColor(1); Grover_xmrkr.SetLineColor(2);
  Grover_yaxis.SetLineColor(1); Grover_ymrkr.SetLineColor(3);
  Grover_zaxis.SetLineColor(1); Grover_zmrkr.SetLineColor(4);
  // rotation matrices
  Double_t ut1 = 90.0, ut2 = 90.0, ut3 = 0.0;
  Double_t up1 =  0.0, up2 = 90.0, up3 = 0.0;
  TRotMatrix Grover_zaxmat("Grover_zaxmat","Grover_zaxmat",
			   ut1,       up1,
			   ut2,       up2,
			   ut3,       up3);
  TRotMatrix Grover_xaxmat("Grover_xaxmat","Grover_xaxmat",
			   ut1+90.0,  up1,
			   ut2,       up2,
			   ut3+90.0,  up3);
  TRotMatrix Grover_yaxmat("Grover_yaxmat","Grover_yaxmat",
			   ut1+90.0,  up1,
			   ut2,       up2+90.0,
			   ut3+90.0,  up3+90.0);
  if (drawaxes) {
    TNode* Grover_node2 = new TNode("","",&Grover_zaxis,0,0,0,&Grover_zaxmat);
    TNode* Grover_node3 = new TNode("","",&Grover_xaxis,0,0,0,&Grover_xaxmat);
    TNode* Grover_node4 = new TNode("","",&Grover_yaxis,0,0,0,&Grover_yaxmat);
    TNode* Grover_node5 = new TNode("","",&Grover_zmrkr,0,0,ZAXISLENGTH*SCALE/2.0,&Grover_zaxmat);
    TNode* Grover_node6 = new TNode("","",&Grover_xmrkr,XAXISLENGTH*SCALE/2.0,0,0,&Grover_xaxmat);
    TNode* Grover_node7 = new TNode("","",&Grover_ymrkr,0,YAXISLENGTH*SCALE/2.0,0,&Grover_yaxmat);
  }


  
  // 6) Volumes for target wires
  int nodecounter = 0; // number of nodes
  char mname[40], sname[40], nname[40]; // matrix,shape,node names
  if (Wires != 0 && drawwires) {
    for (GroverWireList::iterator iwire = Wires->begin();
	 iwire != Wires->end(); ++iwire) {
      ++nodecounter;
      float  x  = iwire->X(),      y = iwire->Y(),      z = iwire->Z();
      float dx  = iwire->SizeX(), dy = iwire->SizeY(), dz = iwire->SizeZ();
      if (iwire->Number() == 2) {
	float swap = dy; dy=dz; dz=swap; // dirty hack for MC
      }
      sprintf(mname,"matrix%d",nodecounter);
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",dx*SCALE/2.0,dy*SCALE/2.0,dz*SCALE/2.0);
      shape->SetLineColor(TARGETCOLOUR);
      TRotMatrix *matrix = new TRotMatrix(mname,mname,ut1,up1,ut2,up2,ut3,up3);
      TNode *node = new TNode(nname,nname,shape,
			      x*SCALE,y*SCALE,z*SCALE,matrix);        
    }
  }





  // 7) Volumes for primary vertices
  if (Primaries != 0 && drawprim) {
    for (GroverVertexList::iterator iprim = Primaries->begin();
	 iprim != Primaries->end(); ++iprim) {
      ++nodecounter;
      float x = iprim->X(), y = iprim->Y(), z = iprim->Z();
      float dx  = sqrt(fabs(iprim->Cov(0)));
      float dy  = sqrt(fabs(iprim->Cov(2)));
      float dz  = sqrt(fabs(iprim->Cov(5)));
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx*nsig*SCALE,
			       dy*nsig*SCALE,
			       dz*nsig*SCALE);
      shape->SetLineColor(PRIMARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x*SCALE,y*SCALE,z*SCALE,
			      &Grover_zaxmat);        
    }
  }





  // 8) Volumes for MC primary vertices
  if (MCPrimaries != 0 && drawmcprim) {
    for (GroverVertexList::iterator iprim = MCPrimaries->begin();
	 iprim != MCPrimaries->end(); ++iprim) {
      ++nodecounter;
      float x = iprim->X(), y = iprim->Y(), z = iprim->Z();
      float dx  = 0.005; // perhaps a little sphere instead?
      float dy  = 0.005;
      float dz  = 0.005;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx*SCALE,
			       dy*SCALE,
			       dz*SCALE);
      shape->SetLineColor(MCPRIMARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x*SCALE,y*SCALE,z*SCALE,
			      &Grover_zaxmat);        
    }
  }




  // 9) Volumes for Primary tracks
  if (Primaries != 0 && drawprimtracks) {
    for (GroverVertexList::iterator iprim = Primaries->begin();
         iprim != Primaries->end(); ++iprim) {
      for (GroverParticleList::iterator ipar = iprim->particles.begin();
           ipar != iprim->particles.end(); ++ipar) {
	float p  = ipar->Momentum();
	float px = p*sin(atan(ipar->BeginSlopeX()));
        float py = p*sin(atan(ipar->BeginSlopeY())); 
	float pt = sqrt(px*px + py*py);
	if (ipar->Momentum() < PCUT ||
	    pt < PTCUT ||
	    ipar->Chi2() > CHI2CUT ||
	    ipar->Nhit() < MINNH) continue; 
        ++nodecounter;
        GroverParticle temp = *ipar;
        temp.ResetParameters(); 
        temp.PropagateBeginTo(iprim->Z());
        float x = temp.BeginX(), y = temp.BeginY(), z = temp.BeginZ();
        float dx2  = temp.BeginCov(0);
        float dy2  = temp.BeginCov(2);
        float dz   = PRIMARYTRACKLENGTH;//fabs(ZAXISLENGTH - z);
        float d    = sqrt(fabs(dx2+dy2));
        sprintf(sname,"shape%d",nodecounter);
        sprintf(nname,"node%d",nodecounter);
        TTUBE *shape = new TTUBE(sname,sname,"void",
                                 0.0, d*SCALE, dz*SCALE/2.0);
        shape->SetLineColor(PRIMARYTRACKCOLOUR);
        shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
        float tx = temp.BeginSlopeX();
        float ty = temp.BeginSlopeY();
	double theta[3]={0.0, 0.0, 0.0};
	double phi[3]={0.0, 0.0, 0.0};
	ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]);
	TRotMatrix *matrix = 
	  new TRotMatrix(mname,mname,
			 theta[0], phi[0],
			 theta[1], phi[1],      
			 theta[2], phi[2]);     
	TNode *node = new TNode(nname,nname,shape,
				x*SCALE,y*SCALE,z*SCALE,
				matrix);        
      }
    }
  }



  // 10) Volumes for secondary vertices
  if (Secondaries != 0 && drawsec) {
    for (GroverVertexList::iterator isec = Secondaries->begin();
	 isec != Secondaries->end(); ++isec) {
      ++nodecounter;
      float x = isec->X(), y = isec->Y(), z = isec->Z();
      float dx  = sqrt(fabs(isec->Cov(0)));
      float dy  = sqrt(fabs(isec->Cov(2)));
      float dz  = sqrt(fabs(isec->Cov(5)));
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx*nsig*SCALE,
			       dy*nsig*SCALE,
			       dz*nsig*SCALE);
      shape->SetLineColor(SECONDARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x*SCALE,y*SCALE,z*SCALE,
			      &Grover_zaxmat); 
    }
  }



  // 11) Volumes for MC secondary vertices
  if (MCSecondaries != 0 && drawmcsec) {
    for (GroverVertexList::iterator isec = MCSecondaries->begin();
	 isec != MCSecondaries->end(); ++isec) {
      ++nodecounter;
      float x = isec->X(), y = isec->Y(), z = isec->Z();
      float dx  = 0.005; // perhaps a little sphere instead?
      float dy  = 0.005;
      float dz  = 0.005;
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TBRIK *shape = new TBRIK(sname,sname,"void",
			       dx*SCALE,
			       dy*SCALE,
			       dz*SCALE);
      shape->SetLineColor(MCSECONDARYCOLOUR);
      TNode *node = new TNode(nname,nname,shape,
			      x*SCALE,y*SCALE,z*SCALE,
			      &Grover_zaxmat);        
    }
  }




  // 12) Volumes for secondary tracks
  if (Secondaries != 0 && drawsectracks) {
    for (GroverVertexList::iterator isec = Secondaries->begin();
	 isec != Secondaries->end(); ++isec) {
      for (GroverParticleList::iterator ipar = isec->particles.begin();
	   ipar != isec->particles.end(); ++ipar) {
        float p  = ipar->Momentum();
        float px = p*sin(atan(ipar->BeginSlopeX()));
	float py = p*sin(atan(ipar->BeginSlopeY()));
        float pt = sqrt(px*px + py*py);
	if (ipar->Momentum() < PCUT ||
	    pt < PTCUT ||
	    ipar->Chi2() > CHI2CUT ||
	    ipar->Nhit() < MINNH) continue;
	++nodecounter;
	GroverParticle temp = *ipar;
        temp.ResetParameters(); 
        temp.PropagateBeginTo(isec->Z());
	float x = temp.BeginX(), y = temp.BeginY(), z = temp.BeginZ();
        float dx2  = temp.BeginCov(0);
        float dy2  = temp.BeginCov(2);
	float dz   = SECONDARYTRACKLENGTH; //fabs(ZAXISLENGTH - z);
	float d    = sqrt(fabs(dx2+dy2));
	sprintf(sname,"shape%d",nodecounter);
	sprintf(nname,"node%d",nodecounter);
	TTUBE *shape = new TTUBE(sname,sname,"void",
				 0.0, d*SCALE, dz*SCALE/2.0);
	shape->SetLineColor(SECONDARYTRACKCOLOUR);
	shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
	float tx = temp.BeginSlopeX();
        float ty = temp.BeginSlopeY();
	double theta[3]={0.0, 0.0, 0.0};
	double phi[3]={0.0, 0.0, 0.0};
	ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]); 
	TRotMatrix *matrix = 
	  new TRotMatrix(mname,mname,
			 theta[0], phi[0],
			 theta[1], phi[1],      
			 theta[2], phi[2]);     
	TNode *node = new TNode(nname,nname,shape,
				x*SCALE,y*SCALE,z*SCALE,
				matrix);        
      }
    }
  }




  // 13) Volumes for unused tracks
  if (Particles != 0 && drawunusedtracks) {
    for (GroverParticleList::iterator ipar = Particles->begin();
	 ipar != Particles->end(); ++ipar) {
      if (ipar->UsedInSecondary() != 0 ||
	  ipar->UsedInPrimary()   != 0   ) continue;
      float p  = ipar->Momentum();
      float px = p*sin(atan(ipar->BeginSlopeX()));
      float py = p*sin(atan(ipar->BeginSlopeY()));
      float pt = sqrt(px*px + py*py);
      if (ipar->Momentum() < PCUT ||
	  pt < PTCUT ||
	  ipar->Chi2() > CHI2CUT ||
	  ipar->Nhit() < MINNH) continue; 
      ++nodecounter;
      ipar->PropagateBeginTo(-3.5);
      float x = ipar->BeginX(), y = ipar->BeginY(), z = ipar->BeginZ();
      float dx2  = ipar->BeginCov(0);
      float dy2  = ipar->BeginCov(2);
      float dz   = UNUSEDTRACKLENGTH;//fabs(ZAXISLENGTH - z);
      float d    = sqrt(fabs(dx2+dy2));
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TTUBE *shape = new TTUBE(sname,sname,"void",
			       0.0, d*SCALE, dz*SCALE/2.0);
      shape->SetLineColor(UNUSEDTRACKCOLOUR);
      shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
      float tx = ipar->BeginSlopeX();
      float ty = ipar->BeginSlopeY();       
      double theta[3]={0.0, 0.0, 0.0};
      double phi[3]={0.0, 0.0, 0.0};
      ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]);
      TRotMatrix *matrix = 
	new TRotMatrix(mname,mname,
		       theta[0], phi[0],
		       theta[1], phi[1],      
		       theta[2], phi[2]);     
      TNode *node = new TNode(nname,nname,shape,
			      x*SCALE,y*SCALE,z*SCALE,
			      matrix);        
    }
  }
  
  
  


  // 14) Volumes for special tracks
  if (Specials != 0 && drawspecialtracks) {
    for (GroverParticleList::iterator ipar = Specials->begin();
	 ipar != Specials->end(); ++ipar) {
      ++nodecounter;
      float x = ipar->BeginX(), y = ipar->BeginY(), z = ipar->BeginZ();
      float dx2  = ipar->BeginCov(0);
      float dy2  = ipar->BeginCov(2);
      float dz   = SPECIALTRACKLENGTH;//fabs(ZAXISLENGTH - z);
      float d    = sqrt(fabs(dx2+dy2));
      sprintf(sname,"shape%d",nodecounter);
      sprintf(nname,"node%d",nodecounter);
      TTUBE *shape = new TTUBE(sname,sname,"void",
			       0.0, d*SCALE, dz*SCALE/2.0);
      shape->SetLineColor(SPECIALTRACKCOLOUR);
      shape->SetNumberOfDivisions(TRACKSMOOTHNESS);
      float tx = ipar->BeginSlopeX();
      float ty = ipar->BeginSlopeY();       
      double theta[3]={0.0, 0.0, 0.0};
      double phi[3]={0.0, 0.0, 0.0};
      ConvertAngles(tx,ty,theta[0],phi[0],theta[1],phi[1],theta[2],phi[2]);
      TRotMatrix *matrix = 
	new TRotMatrix(mname,mname,
		       theta[0], phi[0],
		       theta[1], phi[1],      
		       theta[2], phi[2]);     
      TNode *node = new TNode(nname,nname,shape,
			      x*SCALE,y*SCALE,z*SCALE,
			      matrix);        
    }
  }





  // 15) DRAW IT !!!
  MotherVolume.Draw();
  GroverCanvas->Update();
  if (verbose) cout << "*** GroverDraw3D: Mother volume has "
		    << nodecounter 
		    << " TNode daughters" << endl;
  if (cuts->TDX3D == 1) {
    cout << "*** GroverDraw3D: type 'm' in X3D window for menu"
	 << endl <<"                       'q' to quit and return to ARTE" 
	 << endl;
    GroverCanvas->x3d();
  }




  firstcall = false;
  return 0;
}






////////////////////////////////////////////////////////////////////////



int ConvertAngles(float tx, float ty, 
		  double& theta1, double& phi1,
		  double& theta2, double& phi2,
		  double& theta3, double& phi3) 
{
  //
  // Converts tx,ty to the six Geant angles
  // 
  // S.Schaller 20/10/00
  //

  double a[3][3];
  double theta[3] = {0.0, 0.0, 0.0};
  double phi[3]   = {0.0, 0.0, 0.0};
  int i;         
  const double raddeg = 57.2957795131;        // 1 radian in deg

  double Theta = atan2(sqrt(tx*tx+ty*ty),1.0); // polar angle in radians
  double Phi   = atan2(ty,tx);                 // azimuthal angle in radians

  double ct = cos(Theta);
  double st = sin(Theta);
  double cp = cos(Phi);
  double sp = sin(Phi);
  
  // Rotation Matrix 
  a[0][0] = ct*cp;
  a[0][1] = ct*sp;
  a[0][2] = -1.0*st;
  
  a[1][0] = -1.0*sp;
  a[1][1] = cp;
  a[1][2] = 0;
  
  a[2][0] = st*cp;
  a[2][1] = st*sp;
  a[2][2] = ct; 

  for (i=0; i<3; i++) {  
    theta[i] = raddeg*atan2(sqrt(a[i][0]*a[i][0] +
    	                         a[i][1]*a[i][1]  ),
			         a[i][2]);
    phi[i]   = raddeg*atan2(a[i][1],
			    a[i][0]);  
  }

  theta1 = theta[0];
  theta2 = theta[1];
  theta3 = theta[2];
  phi1   = phi[0];
  phi2   = phi[1];
  phi3   = phi[2];

  return 0;
}


/////////////////////////////////////////////////////////////////////////


// A nice and easy way, since we can't have it in the kuip interface because
// Root won't run on the fourth level farm.
int GroverDraw3D(void)
{
  cout <<"*** GroverDraw3D: ggX sizes: prim,sec,par,mcprim,mcsec,wires,specials= "
       << ggPrimaries.size() << " "
       << ggSecondaries.size() << " "
       << ggParticles.size() << " "
       << ggMCPrimaries.size() << " "
       << ggMCSecondaries.size() << " "
       << ggWires.size() << " "
       << ggSpecialParticles.size() << " "
       << endl;
  return GroverDraw3D(&ggParticles, &ggWires, &ggPrimaries, &ggSecondaries, 
		      &ggMCPrimaries, &ggMCSecondaries, &ggSpecialParticles, 
		      &ggCuts);
}
