#include "arte/arte.hh"
#include "grover/higz.h"
#include "grover/grover.hh" // - class defs, f.prototypes and global variables

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// ------------------------------------------------------------------------
// 00-10-01  M.Sang   added GroverEvent(void) as default
// 00-10-26  M.Sang   declare global variables for Arte implementation
// 00-10-26  M.Sang   Major revision to structure and safety checks
// 00-11-01  M.Sang   RTAR writing implemented

// declare global Grover variables
GroverParticleList ggParticles;
GroverWireList     ggWires;
GroverVertexList   ggPrimaries;
GroverVertexList   ggSecondaries;
GroverVertexList   ggMCPrimaries;
GroverVertexList   ggMCSecondaries;
GroverParticleList ggSpecialParticles;
GroverCuts         ggCuts;


int GroverEvent(void) {

  /////////////////////////////////////////////////////////////////
  //
  // Routine to reconstruct all primary vertices in an event
  // and...
  //
  // Under development. This is the routine that users will usually
  // call. When secondary (J/psi+anything and/or anything) 
  // finding is stable this will go here as well.
  //
  // M.Sang 05 June 00
  //
  /////////////////////////////////////////////////////////////////

  
  // Counter for total number of calls 
  static int total_ncalls = 0;
  ++total_ncalls;
  bool firstcall = (total_ncalls == 1);
  if (firstcall) cout << "*** First call of GroverEvent() from Grover v."
		      << GroverVersion << endl;
  int error = 0;




  // Counter for number of calls in THIS event
  ArtePointer<EVHD> evhd = *(ArteTable<EVHD>::begin());
  static int oldrun, oldevt;
  static int ncalls_thisevt;
  bool duplicall = false;  
  if (firstcall) {
    oldrun = evhd->run; oldevt = evhd->evt; 
    duplicall = false;
    ncalls_thisevt = 1;
  }
  else {
    if (evhd->run == oldrun && evhd->evt==oldevt) {
      duplicall = true; // multiple calls!
      ++ncalls_thisevt;
    }
    else { // first call on this evt
      duplicall = false;
      ncalls_thisevt = 1;
    }
    oldrun = evhd->run; oldevt = evhd->evt; // for next time
  }




  // Code steering
  // 1) Fill track banks?
  bool fill_tracks      = false;
  if ( (!duplicall)                           || 
       (ggParticles.empty())                  || 
       (duplicall && ggCuts.GERefillTracks==1)  ) {
    ggParticles.clear();
    fill_tracks = true;
  }
  // 2) Find primaries?
  bool find_primaries   = false;
  if ( (!duplicall && ggCuts.GEFindPrimaries==1) ||
       ( duplicall && ggCuts.GERedoPrimaries==1)   ) {
    ggPrimaries.clear();
    find_primaries = true;
  }
  // 3) Find Secondaries?
  bool find_secondaries = false;
  if ( (!duplicall && ggCuts.GEFindSecondaries==1) ||
       ( duplicall && ggCuts.GERedoSecondaries==1)   ) {
    // compatibility check
    if (fill_tracks && !find_primaries) {
      cout << "*** GroverEvent: WARNING - secondary vertexing impossible, "
	      "request ignored!"<< endl;
      find_secondaries = false;
      if (ggCuts.GERedoSecondaries==1) 
	ggSecondaries.clear(); // hmmm... not sure about this
    }
    else {
      ggSecondaries.clear();
      find_secondaries = true;
    }
  }
  // 4) Use wire following mechanism or read positions from RTAR?
  const static bool wire_following = (ggCuts.GPWireFollowing==1);
  if (wire_following && 
      (firstcall||evhd->run!=oldrun) && 
      !ArteTable<GTAR>::booked()) return -1; // Cannot continue
  // 5) Printout?
  bool verbose = (ggCuts.Verbose == 1);
  // 6) Do some evaluation - efficiency estimation, etc?
  bool verona = (ggCuts.GPEActivate == 1);
  // 7) Which package?
  bool usingGrover = false;
  bool usingVT     = false;
  /*
  if (ggCuts.VertexingPackage == 1) {
    usingGrover = true;  usingVT = false;
  }
  else if (ggCuts.VertexingPackage == 2) {
    usingGrover = true; usingVT = true; // Grover needed for wire alignment
  }
  else {
    cout << "*** GroverEvent: ERROR - unknown vertexing package ID = "
	 << ggCuts.VertexingPackage << endl;
    return -1;
  }
  */
  usingGrover = true; usingVT = false; // no option in this version!
  // 8) Are we using RSEGs? Do we want them to be refitted with (optionally
  //    smeared) MC momentum?
  static bool rseg_changed = false;
  bool refitting_rsegs = false;
  if (!duplicall) rseg_changed = false;
  if (ggCuts.GERefitRSEGTable==1 && 
      !rseg_changed) { 
    if (ArteTable<MCIN>::booked()) refitting_rsegs = true;
    else {
      if (total_ncalls < 25) 
	cout << "*** GroverEvent: WARNING - real data!"
		" cannot refit VDS RSEGs using MC momentum" << endl;
    }
  }
  // 9) Are we writing out the RTAR table?
  bool write_rtar = (ggCuts.GEWriteRTAR==1 && wire_following);


  // Get wires 
  if (!duplicall) {
    if (wire_following) {
      if (firstcall || evhd->run!=oldrun) { 
	// new wires, please
	ggCuts.GPNewWires=1;
	ggWires.clear();
	GroverFillFromGTAR(&ggWires);
      }
      else ggCuts.GPNewWires = 0; // continue as normal
    }
    else { // no wire following - take directly from GTAR or RTAR tables
      ggCuts.GPNewWires=0;
      ggWires.clear();
      if (!ArteTable<MCIN>::booked() &&
	  ArteTable<RTAR>::booked()   ) 
	GroverFillFromRTAR(&ggWires);
      else GroverFillFromGTAR(&ggWires);
    }
    if (ggWires.empty()) {
      if (total_ncalls < 25)
	cout << "*** GroverEvent: ERROR - no target wires" << endl;
      return -1;
    }
    else for (GroverWireList::iterator iwire = ggWires.begin();
	      iwire != ggWires.end(); ++iwire) iwire->GlobalToLocal();
  }
  


  if (fill_tracks) {
    if (refitting_rsegs) {
      if (verbose) 
	cout << "*** GroverEvent: Refitting VDS RSEGs with MC momenta" << endl;
      int err = GroverRefitRSEG(&ggCuts);
      rseg_changed = true;
      if (err != 0 && total_ncalls < 25) 
	cout << "*** GroverEvent: WARNING - error in rseg refitting" <<endl;
    }

    // Fill particle list from RSEG or RTRA or Rover default (changes with time)
    switch(ggCuts.GETrackChoice) {
    case 1: {
      GroverFillLikeRover(&ggParticles);
      break;
    }
    case 2: {
      GroverFillFromRSEG(&ggParticles);
      break;
    }
    case 3: {
      GroverFillFromRTRA(&ggParticles);
      break;
    }
    default: {
      GroverFillLikeRover(&ggParticles);
      if (total_ncalls<25) 
	cout << "*** GroverEvent: WARNING - unknown track selection"
	        " requested -> using default!" << endl;
      break;
    }
    }
    if (ggParticles.empty()) {
      if (verbose) {
	cout << "*** GroverEvent: WARNING - no ";
	switch(ggCuts.GETrackChoice) {
	case 1: {
	  cout << "particles found using standard selection" << endl;
	  break;
	}
	case 2: {
	  cout << "VDS segments in RSEG" << endl;
	  break;
	}
	case 3: {
	  cout << "tracks in RTRA" << endl;
	  break;
	}
	default: {
	  cout << "particles found using standard selection" << endl;
	  break;
	}
	}
	error -= 1;
      }
    }
    // Move Particles to VDS local coordinates
    for (GroverParticleList::iterator ipar = ggParticles.begin();
	 ipar != ggParticles.end(); ++ipar) ipar->GlobalToLocal();
  }






  // Call primary finder(s)
  int flag = 0;
  if (find_primaries) {
    if (firstcall || verbose) 
      cout << "*** GroverEvent: finding primaries" <<endl;
    flag = GroverPrimaries(&ggParticles, &ggWires, &ggCuts, &ggPrimaries);
    if (flag != 0) {
      if (total_ncalls<25 || verbose) 
	cout << "*** GroverEvent: WARNING - GroverPrimaries error = " 
	     << flag << endl;
      error -= 4;;
    }
    if (usingVT) {
      //rPrimaries.clear();
      cout << "*** GroverEvent: ERROR - VT disabled" << endl;
      //flag = GroverVTPrimaries(&Particles, &Wires, &cuts, &rPrimaries);
      if (flag != 0) {
	if (verbose)
	  cout << "*** GroverEvent: WARNING - GroverVTPrimaries error = " 
	       << flag << endl;
	error -= 8;;
      }
    }
  }





  // Get MC vertices if required
  GroverVertexList* mcverts = 0;
  if (verona || ggCuts.GroverDraw==1 || ggCuts.GroverDraw3D==1) {
    if (ArteTable<MCIN>::booked()) {
      ggMCPrimaries.clear();
      ggMCSecondaries.clear();
      flag = GroverFillFromMVER(&ggMCPrimaries, 1);
      if (flag != 0) {
	if (verbose) 
	  cout << "*** GroverEvent: WARNING - couldn't get "
	    "MC primary vertices" << endl;
	error -= 16;
      }
      else mcverts = &ggMCPrimaries;
    }
  }




  // Calculate primary efficiencies, etc...
  if (verona) {
    cout << "*** GroverEvent: WARNING - GroverPrimEval disabled in this version" << endl;
    //flag = GroverPrimEval(&ggPrimaries,mcverts,&ggCuts);
    flag = -1;
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - GroverPrimEval failed" << endl;
      error -= 32;
    }
  }
  






  // Find Secondaries
  if (find_secondaries) {
    if (firstcall || verbose) 
      cout << "*** GroverEvent: finding secondaries" <<endl;
    flag = GroverSecondaries(&ggParticles, &ggWires, &ggCuts, 
			     &ggPrimaries, &ggSecondaries);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - GroverSecondaries failed" << endl;
      error -= 64;
    }
  }








  // Fill RVER
  if (ggCuts.GEWriteRVER==1) {
    // Clear RVER 
    ArteTable<RVER>::clear();
    flag = GroverAddToRVER(&ggPrimaries, 1);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - couldn't write "
	  "primary vertices into RVER" 
	     << endl;
      error -= 128;
    }
    flag = GroverAddToRVER(&ggSecondaries, 2);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - couldn't write "
	  "secondary vertices into RVER" 
	     << endl;
      error -= 256;
    }
  }



  // Write RTAR
  if (write_rtar) {
    float dxa[9]={0.2};
    float dya[9]={0.2};
    float dza[9]={0.2};
    for (ArteTable<GTAR>::iterator igt = ArteTable<GTAR>::begin();
	 igt != ArteTable<GTAR>::end(); ++igt) {
      int wn = igt->wnr;
      if (wn<0||wn>8) continue;
      dxa[wn] = igt->dx;
      dya[wn] = igt->dy;
      dza[wn] = igt->dz;
    }
    for (int i=0; i<9; i++) {
      if (wire_zsize[i] <=0.0) continue;
      float oldpos[5]={wire_xpos[i],wire_ypos[i],wire_zpos[i],0.0,0.0};
      float newpos[5]={0.0};
      acLocalToGlobal(oldpos,newpos,1);
      RTAR rtar;
      rtar.x  = newpos[0]; 
      rtar.y  = newpos[1]; 
      rtar.z  = newpos[2];
      rtar.sx = rtar.dx = wire_xsize[i]; 
      rtar.sy = rtar.dy = wire_ysize[i];
      rtar.sz = rtar.dz = wire_zsize[i];
      rtar.dx = dxa[i];
      rtar.dy = dya[i];
      rtar.dz = dza[i];
      rtar.wnr=i;
      ArteTable<RTAR>::push_back(rtar);
    }
  }

  
  if (verbose) cout <<"*** ggX sizes: prim,sec,par,mcprim,mcsec,wires,specials"
		    << ggPrimaries.size() << " "
		    << ggSecondaries.size() << " "
		    << ggParticles.size() << " "
		    << ggMCPrimaries.size() << " "
		    << ggMCSecondaries.size() << " "
		    << ggWires.size() << " "
		    << ggSpecialParticles.size() << " "
		    << endl;
  return error;
}



/////////////////////////////////////////////
int GroverEvent(GroverCuts* usercuts)
{ 
  // hack to emulate old behaviour
  // M Sang 00-10-26

  // default
  GroverCuts temp = ggCuts;              // memberwise copy 
  if (usercuts != 0) ggCuts = *usercuts; // -"-
  int r = GroverEvent();
  if (usercuts != 0) ggCuts = temp;      // -"-
  return r;
}
