#include "arte/arte.hh"
#include "grover/grover.hh"
//#include "TRandom.h"       // ROOT random number generators
#include "arte/Chrono.hh"   // A timer

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 00-09-20  Max      Skip very short MTRAs for improved speed
// 00-11-01  Max      Use cernlib random number gen instead of ROOT
// 00-11-03  Max      Error in 1/p covariance element found. The bug will
//                    only be present in ARTE-03-08-r2. Copy this file from
//                    r3 (or dev) if you would like to use it. :-(

extern "C" { // for the *linker's* benefit!
  void CatsRefit(ArtePointer<RSEG>);
}

extern "C" {
  extern void rnorml_(float*, int*);
}

void ReinitializeRSEG(ArtePointer<RSEG>);


int GroverRefitRSEG(GroverCuts* usercuts = 0)
{

  /////////////////////////////////////////////////////////////////////
  //
  // Use momenta from MTRA to refit all VDS RSEGs using CatsRefit.
  // Momenta can be smeared via cuts
  //
  // M.Sang 7 Sep 00
  //
  /////////////////////////////////////////////////////////////////////

  const int MAX_RSEGS = 500;

  if (!ArteTable<MCIN>::booked()) {
    cout << "*** GroverRefitRSEG: ERROR - not Monte Carlo data!" <<endl;
    return -1;
  }


  // error flag
  int errorflag = 0;


  // cuts
  static GroverCuts temp;
  GroverCuts* cuts = &temp;
  if (usercuts != 0) cuts = usercuts;
  // To smear or not to smear?
  static bool  smearing    = (cuts->CheatSmearing == 1);
  // constant fractional error (if smearing)
  static float smear_const =  cuts->CheatSmearConst;
  // fractional error which is proportional to momentum (if smearing)
  static float smear_var   =  cuts->CheatSmearVar;   


  // just some variables
  static ArtePointer<RSEG> rsegs[MAX_RSEGS];
  int i = 0; 
  //static TRandom* myRandom = new TRandom();


  ////Smear the MC momentum
  //float smear;
  //rnorml( &smear );
  ////cout <<"smear = "<< smear<< endl;
  ///mc_mom = mc_mom * (1.0 + smear * sm_var);
  //float invmom = 1.0/mc_mom ;
  ////cout << "Corresponding RSEG " << icat.index() << " gets momentum " 
  ////<< 1.0/invmom << endl;
  //icat->pf=icat->pe=invmom;
  //ReinitializeRSEG(icat);
  //icat->cvf[14] = icat->cve[14] = (sm_var * invmom ) * (sm_var * invmom );
  ////icat->cve[14] = 0.0001*icat->pe*icat->pe;
  ////icat->cvf[14] = 0.0001*icat->pf*icat->pf;


  // loop over mtra and get the rsegs corresponding to them
  for (ArteTable<MTRA>::iterator imtra = ArteTable<MTRA>::begin();
       imtra != ArteTable<MTRA>::end(); ++imtra) {
    if (imtra->leng < 3.0) continue; // MS
    rsegs[i] = RSEGCausedBy(imtra);
    if (rsegs[i] != 0) {
      float mom = sqrt(imtra->px*imtra->px+
		       imtra->py*imtra->py+
		       imtra->pz*imtra->pz );
      float inv_mom = 1.0/mom;
      ReinitializeRSEG(rsegs[i]);
      if (smearing) {
	//
	// (dp/p)^2 = constcon^2 + (p*varcon)^2
	//
	// and the TDR says:
	// ITR: constcon = 0.0040, varcon = 0.00002
	// OTR: constcon = 0.0050, varcon = 0.00005
	// (so roughly constant until high momenta, ~ 50 GeV)
	//
	// expected fractional error (squared)
	float avg_corr_sq =     smear_const*smear_const +
	                    (mom*smear_var)*(mom*smear_var);
	// fractional errors for this track
	//float constcon    =     myRandom->Gaus(0,smear_const);
	float constcon;
	int one=1;
	rnorml_(&constcon,&one);
	constcon = constcon*smear_const;

	//float varcon      = mom*myRandom->Gaus(0,smear_var);
	float varcon;
	rnorml_(&varcon,&one);
	varcon = varcon*mom*smear_var;

	// smear
	float smeared_mom = mom*(1.0+constcon+varcon);
	float smeared_inv_mom = 1.0/smeared_mom;
	rsegs[i]->pf = rsegs[i]->pe = smeared_inv_mom;
	// Variance of inverse momentum = (1/p^2 * fractional error)^2
	// = avg_corr_sq/p^4
	rsegs[i]->cvf[14] = rsegs[i]->cve[14] = avg_corr_sq/(mom*mom);
	// and all off-diagonals to zero
	for (int kk=10;kk<14;kk++) rsegs[i]->cvf[kk] = 0.0;
	if (smeared_inv_mom <= 0.0) {
	  cout << "*** GroverRefitRSEG: ERROR - smeared momentum <= 0.0" << endl;
	  smeared_inv_mom = 0.0;
	  --errorflag;
	}
      }
      else { // not smearing - use momentum directly with small error
	rsegs[i]->pf = rsegs[i]->pe = inv_mom;
	rsegs[i]->cvf[14] = rsegs[i]->cve[14] = 0.00000001;
	for (int kk=10;kk<14;kk++) rsegs[i]->cvf[kk] = 0.0;
      }
      i++;
    }
  }


  // Refit the tracks
  for (int j=0; j<i; j++) CatsRefit(rsegs[j]);

  return errorflag;
}

//////////////////////////////////////////////////////////////////////////


void ReinitializeRSEG(ArtePointer<RSEG> r)
{
  int i;

  for(i=0;i<15;i++) 
    {
      r->cvf[i]=0.0;r->cve[i]=0.0;
    }
  //r->cvf[0]=100.0;r->cvf[2]=100.0;r->cvf[5]=1.0;r->cvf[9]=1.0;r->cvf[14]=0.1;
  //r->cve[0]=100.0;r->cve[2]=100.0;r->cve[5]=1.0;r->cve[9]=1.0;r->cve[14]=0.1;
  // Ivan 30.10.2000
  r->cvf[0]=1.0;r->cvf[2]=1.0;r->cvf[5]=1.0;r->cvf[9]=1.0;r->cvf[14]=0.1;
  r->cve[0]=1.0;r->cve[2]=1.0;r->cve[5]=1.0;r->cve[9]=1.0;r->cve[14]=0.1;
}
