#include <iostream>
#include <algorithm>
#include "arte/arte.hh"
#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
//
//

int GroverUnbiasedSecondaries(GroverParticleList* allParticles,
			      GroverWireList*     allWires,
			      GroverCuts*         usercuts,
			      GroverVertexList*   allPrimaries,
			      GroverVertexList*   allSecondaries) 
{
  /////////////////////////////////////////////////////////////////////
  //
  //                GroverUnbiasedSecondaries
  //
  //  Unbiased search for secondary vertices using geometrical
  //  fitting. 
  //  Author M.Sang
  //
  // Arguments:
  //
  // 1) allParticles   - Any particle flagged as 'not used in primary or secondary'
  //                     will be used to build a secondary candidate 
  // 2) allWires       - Pointer to wires
  // 3) cuts           - If the input pointer=0 or absent, default cuts 
  //                     are used.
  // 4) allPrimaries   - On input this is a list of primary vertices.
  //                     On output they may have been modified! 
  //                     (refit, for example)
  // 5) allSecondaries - On input, this is empty. The output will 
  //                     contain ZERO OR MORE secondaries.
  //
  /////////////////////////////////////////////////////////////////////


  if (allPrimaries == 0) return -1;

  const int ASIZE = 1500;

  // Counter and error flag
  static int EventNumber = 0;
  static bool firstcall = true;
  EventNumber++;
  firstcall = (EventNumber==1);

  // Non-default cuts?
  static GroverCuts cuts;
  if (firstcall && usercuts != 0) cuts = *usercuts; //(Memberwise copy)
  static bool verbose = (cuts.Verbose==1);



  // Initialisation
  if (firstcall || verbose)
    cout << "********* Grover v."
	 << GroverVersion 
	 << ": Finding Unbiased Secondaries **********" << endl;



  // Loop over non-primary particles and find high IP seed tracks
  float seedips[ASIZE] = {0.0};  
  GroverParticleList::iterator seedptr[ASIZE] = {0};
  GroverVertexList::iterator  seedprim[ASIZE] = {0};
  int nseeds = 0;
  static float ipscut = cuts.RSMinSeedIPS;
  static int   min_hits = cuts.RSTrackQualNH;
  float ips=0.0, ipsbest = 1e9; 
  bool d2prim = false;
  GroverVertexList::iterator primbest = 0;
  for (GroverParticleList::iterator ipar = allParticles->begin();
       ipar != allParticles->end(); ++ipar) {
    if (ipar->UsedInPrimary() != 0 || 
	ipar->Nhit() < min_hits) continue;
    ipsbest = 1e9;
    for (GroverVertexList::iterator iprim = allPrimaries->begin();
	 iprim != allPrimaries->end(); ++iprim) {
      ips = ipar->Chi2DistTo(iprim);
      // Keep the chi2 distance to *closest* primary
      if (ips < ipsbest) {
	d2prim = true;
	ipsbest  = ips;
	primbest = iprim;
      }
    }
    if (ipsbest > ipscut && d2prim) { // keep as a seed track
      seedptr[nseeds] = ipar; 
      seedips[nseeds] = ipsbest;
      seedprim[nseeds]= primbest;
      nseeds++;
    }
  }
  if (verbose) cout << "found " << nseeds << " seeds out of " 
       << allParticles->size() << " particles"<< endl;
  if (nseeds == 0) return 0;




  // Make new list of candidate vertices and tracks from which to 
  // build them, then sort in descending order of IPS
  static GroverParticleList SeedTracks;
  static GroverVertexList Candidates;
  SeedTracks.clear(); Candidates.clear(); 
  int nsec = 0;
  for (int i=0; i<nseeds; i++) {
    seedptr[i]->SetSecondaryWeight(seedips[i]);
    seedptr[i]->SetMass(0.140); // <== CHARGED PION MASS!
    SeedTracks.push_back(*(seedptr[i]));
  }
  sort(SeedTracks.begin(), SeedTracks.end(), SortGPBySecWeight);
  for (GroverParticleList::iterator ipar = SeedTracks.begin();
       ipar != SeedTracks.end(); ++ipar)
    if (verbose) cout << "IPS " << ipar->SecondaryWeight() << endl;


  // Loop over seeds until all track addition possibilities 
  // have been exhausted
  float chi2contribcut   = cuts.RSMaxChi2Contrib;
  float vtx_match_cut    = cuts.RSMaxTrkVtxChi2;
  float trk_match_cut    = cuts.RSMaxTrkTrkChi2;
  float two_trk_chi2_cut = cuts.RSMax2TVtxChi2;;
  bool keep_looking = true;
  while (keep_looking) {
    keep_looking = false; // default unless this round is successful

    // loop over unused tracks
    for (GroverParticleList::iterator itrk = SeedTracks.begin();
	 itrk != SeedTracks.end(); ++itrk) {
      if (itrk->UsedInSecondary() != 0) continue;

      bool success = false;

      // ==> 1) Try to add to a candidate vertex
      float best_dist = 1e9; GroverVertexList::iterator best_vtx = 0;
      for (GroverVertexList::iterator isec = Candidates.begin();
	   isec != Candidates.end(); ++isec) {
	// find the best matched secondary to this track
	float dist = itrk->Chi2DistTo(isec);
	if (dist < best_dist) {
	  best_dist = dist;
	  best_vtx  = isec;
	}
      }
      if (best_dist < vtx_match_cut) {
	// add the particle to this vtx and refit. If it's OK, flag as used
	best_vtx->Add(itrk);
	best_vtx->ResetParticles();
	GroverConstrainedFit(best_vtx,-1.0);
	if (best_vtx->particles.back().Chi2() < chi2contribcut) {
	  itrk->SetUsedInSecondary(best_vtx->Index());
	  success = true;
	  keep_looking = true;
	}
	else {
	  // reset everything
	  itrk->SetUsedInSecondary(0);
	  best_vtx->particles.pop_back();
	  best_vtx->ResetParticles();
	  GroverConstrainedFit(best_vtx,0.0);
	  success = false;
	}
      }
      

      if (!success) {
	// ==> 2) Try to make a new candidate out of this and any other used track
	float best_dist = 1e9; GroverParticleList::iterator best_trk = 0;
	for (GroverParticleList::iterator jtrk = SeedTracks.begin();
	     jtrk != SeedTracks.end(); ++jtrk) {
	  if (itrk == jtrk || (jtrk->UsedInSecondary()!=0)) continue;
	  float dist = itrk->Chi2DistTo(jtrk);
	  if (dist < best_dist) {
	    best_dist = dist;
	    best_trk  = jtrk;
	  }
	}
	if (best_dist < trk_match_cut) {
	  // Make a vtx out of these two tracks.
	  // If good, add to list and flag both as used.
	  GroverVertex sec(itrk); sec.Add(best_trk);
	  sec.ResetParticles();
	  GroverConstrainedFit(&sec,0.0);
	  if (sec.Chi2() < two_trk_chi2_cut) {
	    // This is a new secondary candidate
	    sec.SetIndex(++nsec);
	    itrk->SetUsedInSecondary(nsec);
	    best_trk->SetUsedInSecondary(nsec);
	    Candidates.push_back(sec);
	    success = true;
	    keep_looking = true;
	  }
	  else {
	    // reset everything
	    itrk->SetUsedInSecondary(0);
	    best_trk->SetUsedInSecondary(0);
	    success = false;
	  }
	}
      }

    } // looped over all unused tracks...

  }// end of infinite loop






  // Next reject vertices with bad chi-squared per d.o.f. and
  // un-flag particles in global list
  float chi2perdofcut = cuts.RSMaxChi2PDOF;
  if (verbose) cout << "THERE ARE " << Candidates.size() << " CANDIDATES" <<endl;
  for (GroverVertexList::iterator isec = Candidates.begin();
	 isec != Candidates.end(); ++isec) {
    if (isec->Chi2()/isec->Size() > chi2perdofcut) {
      int doomed = isec->Index();
      for (GroverParticleList::iterator ipar = allParticles->begin();
	   ipar != allParticles->end(); ++ipar)
	if (ipar->UsedInSecondary() == doomed) ipar->SetUsedInSecondary(0);
      Candidates.erase(isec--);
      if (verbose) cout << "ERASED A CANDIDATE BECAUSE OF BAD chi2" << endl;
    }
  }



  if (Candidates.empty()) return 0;
 
  

  if (verbose) cout << " FOUND " << Candidates.size() << " SECONDARIES " << endl;
  for (GroverVertexList::iterator iver = Candidates.begin();
       iver != Candidates.end(); ++iver) allSecondaries->push_back(*iver);
 

  if (true) return 0;









  // This will do for now
  //=============================================================================
  // Ignore below this line


  /*

  // Now check primary tracks
  float factor = 0.3333333;// cuts.RSRatioThingy;
  for (GroverVertexList::iterator isec = Seeds.begin();
       isec != Seeds.end(); ++isec) {
    float shortest = 1e9, d2sec=1e9;
    float vvchi2cut = 10;// CHANGE THIS SOME TIME!
    GroverParticleList::iterator particle;
    GroverVertexList::iterator closestsec=0, primary=0;
    for (GroverVertexList::iterator iprim = allPrimaries->begin();
	 iprim != allPrimaries->end(); ++iprim) {
      for (GroverParticleList::iterator ipar = iprim->particles.begin();
	   ipar != iprim->particles.end(); ++ipar) {
	d2sec = ipar->Chi2DistTo(isec);
	if (d2sec < shortest) {
	  shortest   = d2sec;
	  particle   = ipar;
	  closestsec = isec;
	  primary    = iprim;
	}
      }
    }
    float d2prim = factor*particle->Chi2DistTo(primary);
    if (d2sec < d2prim) {
      // No we have a particle which is much closer to 
      // this secondary than to its own primary!
      // See if it fits well - if so, refit primary and 
      // check they're still well separated
      float chi2_b4 = isec->Chi2();
      isec->Add(particle);
      isec->ResetParticles();
      GroverSecondaryFit(isec);
      float chi2 = isec->Chi2();
      if (chi2-chi2_b4 < chi2contribcut) {
	primary->trashcan.clear();
	primary->trashcan.push_back(*particle);
	primary->particles.erase(particle);
	primary->ResetParticles();
	GeoFit(primary);
	if (isec->Chi2DistTo(primary) < vvchi2cut) {
	  primary->particles.push_back(primary->trashcan.back());
	  primary->trashcan.pop_back();
	  primary->ResetParticles();
	  GeoFit(primary);
	  // Now we should be back to normal!
	  cout << " QUICK CHECK ! Chi2 went from "
	       << chi2_b4 << " to " << chi2 
	       << " and back to " << primary->Chi2() << endl;
	}
	else { 
	  // label track as used in secondary and not used in primary
	}
      }
    }
  }



  // Reject vertices with bad chi2 or poor separation from primary
  // Remember to update their labels and, if necessary, return their
  // tracks to the appropriate primary.
  
  cout << " FOUND " << Seeds.size() << " SECONDARIES " << endl;
  for (GroverVertexList::iterator iver = Seeds.begin();
       iver != Seeds.end(); ++iver) allSecondaries->push_back(*iver);


  return 0;

  */

}
