#include <iostream>
#include "grover/DumpFormatter.hh"
#include "grover/grover.hh"

////////////////////////////////////////////////////////////////////////
//  GroverWire class definition
//  ---------------------------
//
//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// ------------------------------------------------------------------------
// 00-10-29  M.Sang   Added constructor for RTAR table
//


GroverWire::GroverWire(void)
  : wnr(-1), 
  x(0.),     y(0.),  z(0.), 
  dx(0.025), dy(1.), dz(0.025),
  shape(1),
  horizontal(0),   vertical(1),
  intarget(false), inlocal(false), inglobal(true),
  ntr(0), nprim(0), nsec(0)
{ // Default constructor is a vertical wire at 0,0,0 with thickness 250*250 um
}

GroverWire::GroverWire(ArtePointer<GTAR> gtar)
  : wnr(gtar->wnr),
    x(gtar->x),  y(gtar->y), z(gtar->z),
    dx(gtar->dx),  dy(gtar->dy), dz(gtar->dz),
    shape(gtar->shap),
    intarget(false), inlocal(false), inglobal(true),
    ntr(0), nprim(0), nsec(0) 
{ // Constructor from GTAR entry
  for (int i=0;i<9;i++) rot[i] = gtar->rot[i];
  horizontal = ( fabs(dx)>fabs(dy) ? 1:0);
  if (horizontal == 0) vertical = 1;
  else vertical = 0;
}

GroverWire::GroverWire(ArtePointer<RTAR> rtar)
  : wnr(rtar->wnr),
    x(rtar->x),  y(rtar->y), z(rtar->z),
    dx(rtar->dx),  dy(rtar->dy), dz(rtar->dz),
    shape(0),
    intarget(false), inlocal(false), inglobal(true),
    ntr(0), nprim(0), nsec(0) 
{ // Constructor from RTAR entry added 29.10.00 MS
  for (int i=0;i<9;i++) rot[i] = -999.;
  horizontal = ( fabs(dx)>fabs(dy) ? 1:0);
  if (horizontal == 0) vertical = 1;
  else vertical = 0;
}

GroverWire::~GroverWire(void) 
{ // default destructor
}

int GroverWire::GlobalToLocal(void)
{
  // Covariance unchanged (not a bad approximation, in fact).
  // Momentum should be re-calculated (uncomment line below if/when
  // the EstimateMomentum() member function becomes available!)
  if (inglobal && !inlocal) {
    float oldpos[5], newpos[5];
    oldpos[0] = x; oldpos[1] = y; oldpos[2] = z; oldpos[3] = 0; oldpos[4] = 0;
    acGlobalToLocal(newpos, oldpos, 1);
    x = newpos[0]; y = newpos[1]; z = newpos[2];
    inlocal = true; inglobal = false;
  }
  return 0;
}

int GroverWire::LocalToGlobal(void)
{
  // Covariance unchanged (not a bad approximation, in fact).
  // Momentum should be re-calculated (uncomment line below if/when
  // the EstimateMomentum() member function becomes available!)
  if (inlocal && !inglobal) {
    float oldpos[5], newpos[5];
    oldpos[0] = x; oldpos[1] = y; oldpos[2] = z; oldpos[3] = 0; oldpos[4] = 0;
    acLocalToGlobal(oldpos, newpos, 1);
    x = newpos[0]; y = newpos[1]; z = newpos[2];
    inlocal = false; inglobal = true;
  }
  return 0;
}  

void GroverWire::Dump(void) { // Dump members to stdout
  cout << "*** Grover: Dump() of GroverWire object" << endl;
  DumpFormatter Nice;
  Nice.DumpClassName("GroverWire");
  Nice.DumpComment("Part of the Grover vertexing package");
  Nice.DumpComment("Please send questions about this class "
		    "to the Grover authors");
  Nice.DumpHeader();
  Nice.DumpPrint("wnr",&wnr);
  Nice.DumpPrint("x",&x);
  Nice.DumpPrint("y",&y);
  Nice.DumpPrint("z",&z);
  Nice.DumpPrint("dx",&dx);
  Nice.DumpPrint("dy",&dy);
  Nice.DumpPrint("dz",&dz);
  Nice.DumpPrint("horizontal",&horizontal);
  Nice.DumpPrint("vertical",&vertical);
  Nice.DumpPrint("rot",rot,9);
  Nice.DumpPrint("intarget",&intarget);
  Nice.DumpPrint("inlocal",&inlocal);
  Nice.DumpPrint("inglobal",&inglobal);
  Nice.DumpPrint("ntr",&ntr);
  Nice.DumpPrint("nprim",&nprim);
  Nice.DumpPrint("nsec",&nsec);
  Nice.DumpFooter();  
}
