
// Boris Fominykh 10.10.2000

// this file contains 4 routines for manipulation over muon HITB hits without toil
// 18.10.2000  - wire number in double layer -  added by B.Fominykh

// muhitnum_(...), muhitpos_(...), muhitatr_(...), muhitdet_(...)


//************************************************************************
// muhitnum_(...) returns the number of muon hits in HITB also as
//                the pointer to first muon hit. All muon hits are
//                in one piece without gaps.
//
// input:  ipr    -  debug print control, print if ipr=1, NO print otherwise
//
// output: Nh     -  number of muon hit in HITB
//         Nf     -  first muon hit number in HITB
//         Ntw    -  number of tube chamber wires in HITB
//         Npw    -  number of pad chamber wires
//         Npd    -  number of pad plates in HITB (one DAQ channel connected with 2 pad plates)
//         Npx    -  number of pixels
//************************************************************************

//************************************************************************
// muhitpos_(...) returns the coordinates of the muon hit in HITB
//
// input:  nht    -  hit number in HITB
//         ipr    -  print control, print if ipr=1, NO print otherwise
//
// output: xh     -  x-coordinate of hit
//         yh     -  y-coordinate of hit
//         zh     -  z-coordinate of hit
//         dxh    -  x-size of hit
//         dyh    -  y-size of hit
//************************************************************************


//************************************************************************
// muhitatr_(...) returns the attributes of the muon hit in HITB
//
// input:  nht    -  hit number in HITB
//         ipr    -  print control, print if ipr=1, NO print otherwise
// output: msl    -  muon superlayer (1-MU1,2-MU2,3-MU3,4-MU4)
//         typ    -  hit type (1 - tube wire, 2 - pad chamber wire, 3 - pad, 4 - pixel)
//         lay    -  layer (= 1,2,3) or quadrant number for pixels
//         ud     -  up (=1) or down (=2)
//         nch    -  chamber number extracted from the chamber name
//         nz     -  chamber side in z-direction, =1 - front side, =2 - rear side
//         nx     -  NxWire (=1-16) or row number for pads (=1,2) or NxPixel (>0)
//         nxd    -  NxWire in double layer (1-32), =0 for pads and pixels
//         ny     -  =0 for wires, NyPad or NyPixel (>0)
//************************************************************************


//************************************************************************
// muhitdet_(...) returns the position/sizes of GEDE element related with the hit.
//
// input:  nht    -  hit number in HITB
//         ipr    -  print control, print if ipr=1, NO print otherwise
// output: csr    -  cosine of rotation angle around y axis
//         snr    -  sinus of rotation angle
//         x1     -  x min of GEDE volume
//         x2     -  x max of GEDE volume
//         y1     -  y min of GEDE volume
//         y2     -  y max of GEDE volume
//         z1     -  z min of GEDE volume
//         z2     -  z max of GEDE volume
//************************************************************************


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream.h>
#include <iomanip.h>

#include "arte/arte.hh"
#include "arte/artefunc.h"
#include "arte/ArteMessage.hh"

#include "murec/MuPixChamber.h"
 static MuPixChamber Pix;

extern "C" {
  int MuNameDecode(char*,int*,int*,int*,int*,int*,int*);


// ********************************************************
void mupixpar_(int* iv,int* nx,int* ny,int* nxc,int* nyc) {
// ********************************************************

  *nx = 0;
  *ny = 0;
  *nxc= 0;
  *nyc= 0;
  ArtePointer<GEDE> gede(*iv);
  ArtePointer<GESL> gesl = gede->gesl;
  int sl  = gesl->nsl-1;
  if(sl < 0 || sl > 3) {
    //    cout << "MuPixPar: wrong sl= " << sl << endl;
    return;
  }
  *nx = int(Pix.NxM[sl]);
  *ny = int(Pix.NyM[sl]);
  *nxc= int(Pix.NxC[sl]);
  *nyc= int(Pix.NyC[sl]);
  return;
}


//*********************************************************
void muhitnum_(int* ipr,int* Nh,int* Nf,int* Ntw,int* Npw,int* Npd,int* Npx) {
//*********************************************************

  *Nh  = 0;
  *Nf  = 0;
  *Ntw = 0;
  *Npw = 0;
  *Npd = 0;
  *Npx = 0;

// total number of HITB hits
  int nhit = ArtePointer<HITB>::nrow();
  if(nhit < 1) return;

  int nhmu = 0;
  int nfmu = 0;
  int ntwm = 0;
  int npwm = 0;
  int npdm = 0;
  int npxm = 0;

// loops over HITB
  int nht=0;
  for (ArteTableIterator<HITB> ih = ArteTable<HITB>::begin(); 
  ih != ArteTable<HITB>::end();
  ++ih) {
    nht+=1;
    ArtePointer<GEDE> gede=(*ih)->gede;
    int cmp = gede->cmp;
    if(cmp == 8 || cmp == 9) {
      ArtePointer<GESL> gesl = gede->gesl;
// muon superlayer
      int ms = gesl->nsl;     // # muon superlayer
// number of muon hits and the first hit position
      nhmu+=1;
      if(nfmu == 0 || nht < nfmu) nfmu=nht;
// hit y coordinate
      float y = (*ih)->y;
// type of hit
      int type=0;
      if(cmp == 8) {
        if(ms < 3) type=1;
        if(ms > 2 && y == 0.) type=2;
        if(ms > 2 && y != 0.) type=3;
      }
      if(cmp == 9) type=4;

      if(type == 1) ntwm+=1; 
      if(type == 2) npwm+=1; 
      if(type == 3) npdm+=1; 
      if(type == 4) npxm+=1; 
    }
  }

  if(*ipr == 1) {
    cout << "MuHitNum: Nhitb " << nhit << " Nmuhit " << nhmu;
    cout << " 1st " << nfmu << " Ntw " << ntwm << " Npw " << npwm;
    cout << " Npd " << npdm << " Npx " << npxm;
    cout << endl;
  }
  if(nhmu > 0) {
    *Nh  = nhmu;
    *Nf  = nfmu;
    *Ntw = ntwm;
    *Npw = npwm;
    *Npd = npdm;
    *Npx = npxm;
  } 
  return;
}


//*********************************************************
void muhitpos_(int* nht,int* ipr,float* x,float* y,float* z,float* dx,float* dy) {
//*********************************************************

  *x = 0.;
  *y = 0.;
  *z = 0.;
  *dx= 0.;
  *dy= 0.;

// total number of hits
  int nhit = ArtePointer<HITB>::nrow();
// check nht
  if(*nht < 0 || *nht > nhit) return;
  ArtePointer<HITB> hit(*nht);
// check that it is the muon hit
  ArtePointer<GEDE> gede=hit->gede;
  int cmp = gede->cmp;
  if(cmp != 8 && cmp != 9) return;
 
// hit coordinates
  *x = hit->x;
  *y = hit->y;
  *z = hit->z;
// hit sizes
  *dx = hit->q;
  *dy = hit->r;

  if(*ipr == 1) {
    cout << " hit " << *nht << " x " << *x << " y " << *y;
    cout << " z " << *z << " dx " << *dx << " dy " << *dy; 
    cout << endl;
  }

  return;
}


//******************************************************************
void muhitatr_(int* nht,int* ipr,int* msl,int* typ,int* lay,int* ud,
               int* nch,int* nz, int* nx,int* nxd, int* ny) {
//******************************************************************

  *msl = 0;
  *typ = 0;
  *lay = 0;
  *ud  = 0;
  *nch = 0;
  *nz  = 0;
  *nx  = 0;
  *nxd = 0;
  *ny  = 0;

// total number of hits
  int nhit = ArtePointer<HITB>::nrow();
// check nht
  if(*nht < 0 || *nht > nhit) return;
 
  ArtePointer<HITB> hit(*nht);
  ArtePointer<GEDE> gede=hit->gede;
  int cmp = gede->cmp;
  if(cmp == 8 || cmp == 9) {
    ArtePointer<GESL> gesl = gede->gesl;
// muon superlayer
    int ms = gesl->nsl;     
// take the chamber name
    char name[10];
    strncpy(name,gede->hwc,9);

// decode the chamber name
    int type = 0;
    int msn = 0;
    int layn= 0;
    int udn = 0;
    int nxn = 0;
    int ncn = 0;
    MuNameDecode(&name[0],&type,&msn,&layn,&udn,&nxn,&ncn);
// check-point for ms=msn

    //    cout << name << " nxn " << nxn;
    //    cout << " ms " << ms << " msn " << msn;
    //    cout << " type " << type << " layn " << layn;
    //    cout << " udn " << udn; 
    //    cout << endl;

    if(ms != msn) return;

// hit coordinates
    float x = hit->x;
    float y = hit->y;
    float z = hit->z;

// define the side
    int sdw=0;
    int sd =0;
    int iv[2];
      iv[0]=0;
        iv[1]=0;
    int iv1 = gesl->fpl;
    int iv2 = iv1 + gesl->npl;
    if(iv2 >= iv1) {
      for ( int ivw=iv1; ivw<iv2+1; ivw++) { 
        ArtePointer<GEDE> gdw(ivw);
        char namw[10] = "         ";
        strncpy(namw,gdw->hwc,9);
        if(!strcmp(name,namw)) {
          iv[sdw]=ivw;
          sdw++;
          float z1w  = gdw->zmin;   // Zmin of chamber
          float z2w  = gdw->zmax;   // Zmax of chamber
          if(z > z1w && z< z2w) sd=sdw;
        }
        if(sdw == 2 && gdw == gede) break;
      }
    }
// check-point for side
    if(sd < 1 || sd > 2) return;

// type of hit
    type=0;
    if(cmp == 8) {
      if(ms < 3) type=1;
      if(ms > 2 && y == 0.) type=2;
      if(ms > 2 && y != 0.) type=3;
    }
    if(cmp == 9) type=4;

    float xd1  = gede->xmin;   // Xmin of chamber
    float xd2  = gede->xmax;   // Xmax of chamber
    float yd1  = gede->ymin;   // Ymin of chamber
    float yd2  = gede->ymax;   // Ymax of chamber
    float zd1  = gede->zmin;   // Zmin of chamber
    float zd2  = gede->zmax;   // Zmax of chamber

    //    cout << name;
    //    cout << " cmp " << cmp << " sd " << sd;
    //    cout << " iv[0] " << iv[0] << " iv[1] " << iv[1];
    //    cout << " zd1 " << zd1 << " z " << z << " zd2 " << zd2 << endl;
    //    cout << " xd1 " << xd1 << " x " << x << " xd2 " << xd2;
    //    cout << " yd1 " << yd1 << " y " << y << " yd2 " << yd2 << endl;

// the case of wires
    if(type < 3) {
      float dist = gede->pit;    // distance between wires
      float xfw  = gede->xfw[0]; // X-position of first wire
      int nw = int((x-xd1-xfw+0.1*dist)/dist)+1;
// wire number in double layer
      int nwd = 0;
      if(iv[0] > 0 && iv[1] > 0) {
        ArtePointer<GEDE> g1w(iv[0]);
        float x1w  = g1w->xmin;
        ArtePointer<GEDE> g2w(iv[1]);
        float x2w  = g2w->xmin;
        if(x1w < x2w) nwd=2*nw+sd-2;
        else          nwd=2*nw-sd+1;

	//        cout << " x1w " << x1w << " x2w " << x2w;
	//        cout << " sd " << sd << " nw " << nw << " nwd " << nwd << endl;

      }

      *msl = ms;
      *typ = type;
      *lay = layn;
      *ud  = udn+1;
      *nch = nxn;
      *nz  = sd;
      *nx  = nw;
      *nxd = nwd;
      *ny  = 0;

      if(*ipr == 1) {
        float z = hit->z;
        cout << " wire in " << name;
        cout << " zd1 " << zd1 << " z " << z << " zd2 " << zd2;
        cout << " xd1 " << xd1 << " x " << x << " xd2 " << xd2 << endl;
        cout << " msl " << *msl << " typ " << *typ;
        cout << " lay " << *lay << " ud " << *ud;
        cout << " nch " << *nch << " nz " << *nz;
        cout << " nx " << *nx << " nxd " << *nxd;
        cout << " ny " << *ny;
        cout << endl;
      }

      return;
    }

// the case of pad
    if(type == 3) {
      float yl=yd2-yd1;
      float yl2=0.01*float(gede->nwi[3]); // y-dist between pads
      float dypd1=gede->xfw[2];           // y-size of first pad
      float dypd2=gede->xfw[3];           // y-size of standard pad
      float alsy =gede->symx - gede->symn;
      float yl1  =(yd2-yd1-alsy)/2.;      // y-dist from edge to pad
      int npmx = gede->nwi[2];            // number of pads
// y-size of last pad
      float dypd3=alsy-dypd1-yl2-(dypd2+yl2)*float(npmx-2);
// row number
      int nrow=0;
      if(x > 0.5*(xd1+xd2)) nrow=1;
// pad  number
      int nyw=0;
      float dy=y-yd1;
      if(udn == 1) dy=yd2-y;
      if(dy < yl1+dypd1) nyw=0;
      if(dy > yl1+dypd1+yl2 && dy < yl-yl1-dypd3-yl2) {
        nyw=int((dy-yl1-dypd1-yl2)/(dypd2+yl2))+1;
      } 
      if(dy > yl-yl1-dypd3-yl2) nyw=npmx-1;

      *msl = ms;
      *typ = type;
      *lay = layn;
      *ud  = udn+1;
      *nch = nxn;
      *nz  = sd;
      *nx  = nrow+1;
      *ny  = nyw+1;

      if(*ipr == 1) {
        float z = hit->z;
        float dxh = hit->q;
        float dyh = hit->r;
        cout << name;
        cout << " pad z " << z << " x " << x << " y " << y;
        cout << " dx " << dxh << " dy " << dyh;
        cout << " msl " << *msl << " typ " << *typ;
        cout << " lay " << *lay << " ud " << *ud;
        cout << " nch " << *nch << " nz " << *nz;
        cout << " nrow " << *nx << " ny " << *ny;
        cout << endl;
      }

      return;
    }
 
// the case of pixels
    if(type == 4) {
      int nxm=0;
      int nym=0;
      int nxc=0;
      int nyc=0;
      mupixpar_(&iv[0],&nxm,&nym,&nxc,&nyc);

      float xed  = gede->xfw[0]; // x-dist from edge to pixels
      float xsz  = gede->xfw[2]; // x size of elemetary pixel
      float yed  = gede->xfw[1]; // y-dist from edge to pixels
      float ysz  = gede->xfw[3]; // y size of elemetary pixel
      int qu=0;
      if(fabs(xd2) > fabs(xd1) && fabs(yd2) > fabs(yd1)) qu=0;
      if(fabs(xd2) < fabs(xd1) && fabs(yd2) < fabs(yd1)) qu=1;
      if(fabs(xd2) < fabs(xd1) && fabs(yd2) > fabs(yd1)) qu=2;
      if(fabs(xd2) > fabs(xd1) && fabs(yd2) < fabs(yd1)) qu=3;
// ny
      float ycell=ysz*float(nyc);
// take the distance between outside edge and point to have
// the uniform structure except the last pixel raw
      float dy=yd2-y;
      if(qu == 1 || qu == 3) dy=y-yd1;
      int nyw=int((dy-yed)/ycell);

// trap for the cases when hit is outside chamber
      if(y < yd1+yed || y > yd2-yed) {
        cout << "MuHitAtr: pix y outside chamber,";
        cout << " msl " << msl << " qu " << qu << " ny " << nyw;
        cout << endl;
        cout << " yd1 " << yd1 << " y " << y << " yd2 " << yd2;
        cout << " yed " << yed;
        cout << " dy-yed " << dy-yed << " ycell " << ycell << endl;
        return;
      }

// nx
      float xcell=xsz*float(nxc);
// take the distance between outside edge and point to have
// the uniform structure except the last pixel raw
      float dx=xd2-x;
      if(qu == 1 || qu == 2) dx=x-xd1;
      int nxw=int((dx-xed)/xcell);

// trap for the cases when hit is outside chamber
      if(x < xd1+xed || x > xd2-xed) {
        cout << "MuHitAtr: pix x outside chamber,";
        cout << " msl " << msl << " qu " << qu << " nx " << nxw;
        cout << endl;
        cout << " xd1 " << xd1 << " x " << x << " xd2 " << xd2;
        cout << " xed " << xed;
        cout << " dx-xed " << dx-xed << " xcell " << xcell << endl;
        return;
      }

      *msl = ms;
      *typ = type;
      *lay = qu+1;
      *ud  = udn+1;
      *nch = nxn;
      *nz  = sd;
      *nx  = nxw+1;
      *ny  = nyw+1;

      if(*ipr == 1) {
        float z   = hit->z;
        float dxh = hit->q;
        float dyh = hit->r;
        cout << name;
        cout << " pixel z " << z << " x " << x << " y " << y;
        cout << " dx " << dxh << " dy " << dyh;
        cout << " msl " << *msl << " typ " << *typ << endl;
        cout << "   quadrant " << *lay;
        cout << " nch " << *nch;
        cout << " nx " << *nx << " ny " << *ny;
        cout << endl;
      }

      return;
    }
  }
  return;
}


//*********************************************************
void muhitdet_(int* nht,int* ipr,float* csr,float* snr,
                                 float* x1,float* x2,
                                 float* y1,float* y2,
                                 float* z1,float* z2) {
//*********************************************************

  *csr= 0.;
  *snr= 0.;
  *x1 = 0.;
  *x2 = 0.;
  *y1 = 0.;
  *y2 = 0.;
  *z1 = 0.;
  *z2 = 0.;

// total number of hits
  int nhit = ArtePointer<HITB>::nrow();
// check nht
  if(*nht < 0 || *nht > nhit) return;
  ArtePointer<HITB> hit(*nht);
// check that it is the muon hit
  ArtePointer<GEDE> gede=hit->gede;
  int cmp = gede->cmp;
  if(cmp != 8 && cmp != 9) return;
 
// detector position
  *csr  = gede->rot[0]; // cosine of rotation angle
  *snr  = gede->rot[1]; // sinus of rotation angle
  *x1   = gede->xmin;   // Xmin of chamber
  *x2   = gede->xmax;   // Xmax of chamber
  *y1   = gede->ymin;   // Ymin of chamber
  *y2   = gede->ymax;   // Ymax of chamber
  *z1   = gede->zmin;   // Zmin of chamber
  *z2   = gede->zmax;   // Zmax of chamber

  if(*ipr == 1) {
    cout << " hit " << *nht << " cos " << *csr << " sin " << *snr;
    cout << " Xmin " << *x1 << " Xmax " << *x2;
    cout << endl;
    cout << "   Zmin " << *z1 << " Zmax " << *z2;
    cout << "   Ymin " << *y1 << " Ymax " << *y2;
    cout << endl;
  }

  return;
}
}


