// 
//  Author:  Rainer Mankel                     Created:  19-Apr-99
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 

// rgzone_
// Main access function for navigation table.
// For a given domain layer, provide all navigation targets (ie possible
// domain layers for the next hit) in a given zone which satisfy given
// constraints. The target domain layer pointers are written into a 
// fixed-length array which must be provided by the calling routine.
// nTarget specifies on input the number of elements already stored in
// this array, the new targets are appended after this position and nTarget
// is increased by the number of new elements.

// Method: if the required zone is available without leaving the present domain
// window, only the target in the present domain window is delivered. 
// Otherwise, the domain window navigation table is applied, and the search
// is continued in each navigation target window. This is done by a recursive
// call of rgzone_ in itself.

// Parameters:
//     dl     - current domain layer from which navigation starts
//     method - navigation method, can be
//                 0 - for 3D propagation
//                 1 - for zero degree (horizontal) projection propagation
//                 2 - for stereo layers (vertical) propagation
//     zone   - logical distance of navigation. zone 1 corresponds
//              to domain layers which can be reached directly, zone 2
//              corresponds to domain layers reachable by traversing
//              an intermediate domain layer without finding a hit, etc.
//              (zone-1) implies the number of assumed faults in sequence
//              in the next propagation step.
//     direction - specifies forward propagation (+1), in +z direction, 
//              or backward (-1) propagation, in -z direction. 
//     minScope - minimal scope value for targets to be included. Each
//              domain has a scope identifier which associates it to a 
//              logical tracker part. A typical assignment may be
//                            scope=1  for vertex detector
//                                  2  for magnet tracker
//                                  3  for pattern tracker
//                                  4  for tc area
//              Thus, the navigation can be restricted to a particular 
//              scope range.
//     maxScope - maximal scope for targets to be included.
//     maxExpand - limits the number of domain window navigation table usages 
//              in the same search tree. Note: values greater than 1 may lead 
//              to duplicate entries in the target list which must be
//              handled by the calling program.
//     maxTarget - maximum capacity of target vector
//     targetVec - array for output of navigation target domain layer pointers.
//     nTarget  - number of elements currently in targetVec.

// Important note: contrary to other rg routines with external array interface,
// rgzone is explicitly designed to work iteratively and cumulating, so
// the array index nTarget is not (!!) set to zero by rgzone!

#include "rang/DomWin.hh"
#include "rang/DomLay.hh"
#include "rang/NavItem.hh"
#include "rang/rgzone.hh"

extern "C" {
  void rgzone_(rangerDomLay* const & dl,// navigation start point (input)
	       const int& method,      // navigation method (input)
	       const int& zone,        // target zone (input)
	       const int& direction,   // navigation direction (input)
	       const int& minScope,    // minimal allowed scope (input)
	       const int& maxScope,    // maximal allowed scope (input)
	       const int& maxExpand,   // maximum number of nav table 
                                       //   expansions (input)
	       const int& maxTarget,   // capacity of targetVec (input)
	       rangerDomLay** targetVec, // array of nav targets (in/out)
	       int& nTarget) {         // current number of targets (in/out) 
    
    if (! dl) return;
    if (nTarget >= maxTarget) return;
    if (zone < 0) return;

    // zone zero means we have already arrived
    if (zone == 0) {
      ++nTarget;
      targetVec[nTarget-1] = dl;
      return;
    }

    // try to resolve in current window using the NEXT pointer
    rangerDomLay* dln = dl->next(method,direction);

    if (dln) {
      // zone one and next layer in same window available: take that one
      if (zone == 1) {
	++nTarget;
	targetVec[nTarget-1] = dln;
	return;
      }
      // higher zone and next layer in same window available: iterative call
      else {
	rgzone_(dln,method,zone-1,
		direction,minScope,maxScope,maxExpand,
		maxTarget,targetVec,nTarget);
      }
    }
    else {
      // no more target layers in this domain window: apply the
      // domain window navigation table for continuation in next domain
      if (dl->win()->getNI()) {
	for (vector<rangerNavItem*>::const_iterator 
	       ni= dl->win()->getNI()->getTargets(direction).begin();
	     ni != dl->win()->getNI()->getTargets(direction).end();
	     ++ni) {
	  if ((*ni)->win()->dom()->sco() >= minScope &&
	      (*ni)->win()->dom()->sco() <= maxScope) {

	    // apply rgzone to the entrance layer of this new window
	    if (! (*ni)->win()->entryDL(direction)) {
	      cout << "rgzone: entrance layer missing!" << endl;
	    }
	    rgzone_((*ni)->win()->entryDL(direction),method,zone,
		    direction,minScope,maxScope,maxExpand-1,
		    maxTarget,targetVec,nTarget);

	    // if no further expansion of domain window navigation table
	    // was allowed: skip all others after the first window
	    if (maxExpand <=0) break;
	  }
	}
      }
    }
  }
}
	  
