
// ****************************************************************************
//
// Routine:      rc99AStandAlone
// Author:       T.Oest/D.Samtleben
//
// Description:  stand alone reconstruction
//
//               reco level 0: no reconstruction
//                          1: hitb only
//                          2: reconstruction only
//                          3: hitb + reconstruction
//
// Modification:  
// drop RICH hits
// recoTARGET,... variables no longer static                        27/05/99 TO
// VDS==SVD for kumac-steering error-message if given subdetector
// is not known                                                     10/06/99 DS
// if at least one reconstruction package of a subdetector
// is called the Rxxx-tables are cleared                            01/07/99 DS
// bug fix: proper termination of input[]
// include CATS: recoVDSmodule
// correct calling sequenz for rc99ATARGET                          26/07/99 TO
// global alignment transformation added                            06/09/99 TO
// RICH hits are no longer removed since they are needed for
// the particle ID                                                  14/09/99 TO 
// move VDS global alignment transformation to rc99AVDS             20/09/99 TO
// filling of GTAR bank moved to acReadTARGET                       24/09/99 TO
// include global VDS alignment transformation                      29/10/99 TO
// ITR added                                                        07/02/00 TO
// use ARTE_mark()                                                  12/05/00 AG
// introduce timing                                                 17/05/00 AG
// timing info into globals                                         18/05/00 AG
// creation of chrono jobs only once per job                        24/05/00 DS
// initialize timers to 0                                           07/06/00 AG
// use more ARTE_mark()                                             08/06/00 AG
// polishing                                                        16/06/00 AG
// split ECAL reco in stand-alone and matching part                 21/07/00 AG
// added debug flag                                                 26/07/00 AG
// initialized timers correctly                                     27/07/00 AG
// improved printout                                                08/08/00 AG
// added more debug printout for OTR and RICH cuts                  25/08/00 AG
// introduced rcRANGER and removed ITR and OTR                      13/11/00 AG
//
// ****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "reco/rc99Areco.hh"
#include "reco/rcflags.hh"
#include "rang/zfpar.hh"
#include "rise/RingSearchAttrib.h"

// timers
float time_rc99AVDS   = 0;
float time_rc99AHPT   = 0;
float time_rcRANGER   = 0;
float time_rc99ARICH  = 0;
float time_rc99ATRD   = 0;
float time_rc99AECAL  = 0;
float time_rc99AECAL2 = 0;
float time_rc99AMUON  = 0;

// selectors
int selVDS   = 1;
int selITR   = 1;
int selHPT   = 1;
int selOTR   = 1;
int selRICH  = 1;
int selTRD   = 1;
int selECAL  = 1;
int selECAL2 = 1;
int selMUON  = 1;

// externals
extern int FATIME;
extern int FADEB;

// prototypes
void acRSEGinGlobalCoordinates( const int CMPnumber );

// ECAL
extern "C" void careevnt1_();
extern "C" void careevnt2_();

void matchRsegCare();
int  rcModSel();

int rcRANGER( int recoITR, int recoOTR );

//********************************************************************
// rc99AStandAlone
//********************************************************************
int
rc99AStandAlone()
{
  // mark
  ARTE_mark( "rc99AStandAlone 0", 0 );
  
  //===================================================================
  //     remove the Rxxx tables (for real data)
  //===================================================================
  
  //------  remove Rxxx-tables
  if( recoVDS  ||
      recoITR  ||
      recoHPT  || 
      recoOTR  ||
      recoRICH || 
      recoTRD  || 
      recoECAL || 
      recoMUON    ){
    ArteTable<RREL>::clear();
    ArteTable<RPNT>::clear();
    ArteTable<RVER>::clear();
    ArteTable<RHIT>::clear();
    ArteTable<RSEG>::clear();
    ArteTable<RTRA>::clear();
    ArteTable<RCCL>::clear();
  }

  //===================================================================
  //     initializations
  //===================================================================
  
  // timers
  static Chrono *cron_rc99AVDS   = 0;
  static Chrono *cron_rc99AHPT   = 0;
  static Chrono *cron_rcRANGER   = 0;
  static Chrono *cron_rc99ARICH  = 0;
  static Chrono *cron_rc99ATRD   = 0;
  static Chrono *cron_rc99AECAL  = 0;
  static Chrono *cron_rc99AECAL2 = 0;
  static Chrono *cron_rc99AMUON  = 0;

  static bool first = 1;
  if( first ){
    first = 0;
    ChronoManager CM;
    CM.instance();
    
    cron_rc99AVDS   = CM.createChrono( "rc99AVDS"   );
    cron_rc99AHPT   = CM.createChrono( "rc99AHPT"   );
    cron_rcRANGER   = CM.createChrono( "rcRANGER"  );
    cron_rc99ARICH  = CM.createChrono( "rc99ARICH"  );
    cron_rc99ATRD   = CM.createChrono( "rc99ATRD"   );
    cron_rc99AECAL  = CM.createChrono( "rc99AECAL"  );
    cron_rc99AECAL2 = CM.createChrono( "rc99AECAL2" );
    cron_rc99AMUON  = CM.createChrono( "rc99AMUON"  );
  }

  // initialization
  int VDSflag    = 0;
  int HPTflag    = 0;
  int RANGERflag = 0;
  int RICHflag   = 0;
  int TRDflag    = 0;
  int ECALflag   = 0;
  int ECALflag2  = 0;
  int MUONflag   = 0;

  //===================================================================
  //     stand-alone modules
  //===================================================================

  // VDS
  time_rc99AVDS = 0;
  if( recoVDS && selVDS ){
    ARTE_mark( "rc99AVDS", 0 );
    cron_rc99AVDS->start();
    VDSflag = rc99AVDS(recoVDS);
    cron_rc99AVDS->stop();
    time_rc99AVDS = cron_rc99AVDS->lastTime();
  }

  // ECAL
  time_rc99AECAL = 0;
  if( recoECAL && selECAL ){
    ARTE_mark( "careevnt1_", 0 );
    cron_rc99AECAL->start();
    careevnt1_();
    cron_rc99AECAL->stop();
    time_rc99AECAL = cron_rc99AECAL->lastTime();
  }

  //==================================================================
  // module selection
  //==================================================================
  ARTE_mark( "rcModSel" , 0 );
  int flagModSel = rcModSel();
  if( FADEB ) cout << " rc99AStandAlone: flagModSel=" << flagModSel << endl;

  // HPT
  time_rc99AHPT = 0;
  if( recoHPT && selHPT ){
    ARTE_mark( "rc99AHPT", 0 );
    cron_rc99AHPT->start();
    HPTflag = rc99AHPT(recoHPT);
    cron_rc99AHPT->stop();
    time_rc99AHPT = cron_rc99AHPT->lastTime();
  }

  // RANGER
  time_rcRANGER = 0;
  if( ( recoITR && selITR ) || ( recoOTR && selOTR ) ){
    if( FADEB ) cout << " rc99AStandAlone: OTR: qpmax=" << zfpar_.qpmax << endl;
    ARTE_mark( "rcRANGER", 0 );
    cron_rcRANGER->start();
    RANGERflag = rcRANGER( recoITR, recoOTR );
    cron_rcRANGER->stop();
    time_rcRANGER = cron_rcRANGER->lastTime();
  }

  // RICH
  time_rc99ARICH = 0;
  if( recoRICH && selRICH ){
    if( FADEB ) cout << " rc99AStandAlone: RICH: " 
		     << "occupancyCut="     << RingSearchAttrib::overwriteOccupancyCut << " "
		     << "seedOccupancyCut=" << RingSearchAttrib::seedOccupancyCut      << endl;
    ARTE_mark( "rc99ARICH", 0 );
    cron_rc99ARICH->start();
    RICHflag = rc99ARICH(recoRICH);
    cron_rc99ARICH->stop();
    time_rc99ARICH = cron_rc99ARICH->lastTime();
  }

  // TRD
  time_rc99ATRD = 0;
  if( recoTRD && selTRD ){
    ARTE_mark( "rc99ATRD", 0 );
    cron_rc99ATRD->start();
    TRDflag = rc99ATRD(recoTRD);
    cron_rc99ATRD->stop();
    time_rc99ATRD = cron_rc99ATRD->lastTime();
  }

  // MUON
  time_rc99AMUON = 0;
  if( recoMUON && selMUON ){
    ARTE_mark( "rc99AMUON", 0 );
    cron_rc99AMUON->start();
    MUONflag = rc99AMUON(recoMUON);
    cron_rc99AMUON->stop();
    time_rc99AMUON = cron_rc99AMUON->lastTime();
  }

  //===================================================================
  //     transformation of segments/RHIT to global coordinates
  //===================================================================  
  ARTE_mark( "acRSEGinGlobalCoordinates", 0 );
  if( recoVDSmodule & 1 ) acRSEGinGlobalCoordinates( 1 );
  if( recoVDSmodule & 2 ) acRSEGinGlobalCoordinates( 65537 );
  if( recoRICH )          acRSEGinGlobalCoordinates( 67108864 );
  
  //===================================================================
  //     ECAL needs tracks and trafo for the 2. reconstruction step
  //===================================================================
  time_rc99AECAL2 = 0;
  if( recoECAL && selECAL2 ){
    ARTE_mark( "rc99AECAL2", 0 );
    cron_rc99AECAL2->start();
    //ECALflag = rc99AECAL(recoECAL); // 1. and 2. step
    //ARTE_mark( "careevnt1_", 0 );
    //careevnt1_();    // ECAL 1
    ARTE_mark( "matchRsegCare", 0 );
    matchRsegCare(); // Select the tracks (RSEG entries) for VDS-OTR
    ARTE_mark( "careevnt2_", 0 );
    careevnt2_();    // ECAL reconstruction using tracks
    cron_rc99AECAL2->stop();
    time_rc99AECAL2 = cron_rc99AECAL2->lastTime();
  }

  //===================================================================
  //     print
  //===================================================================
  if( FATIME ){

    float sum =
      time_rc99AVDS   +
      time_rc99AHPT   +
      time_rcRANGER   +
      time_rc99ARICH  +
      time_rc99ATRD   +
      time_rc99AECAL  +
      time_rc99AECAL2 +
      time_rc99AMUON;
    
    cout << " RECO: "
	 << "vds="   << time_rc99AVDS   << " "
	 << "hpt="   << time_rc99AHPT   << " "
	 << "rang="  << time_rcRANGER   << " "
	 << "rich="  << time_rc99ARICH  << " "
	 << "trd="   << time_rc99ATRD   << " "
	 << "ecal1=" << time_rc99AECAL  << " "
	 << "ecal2=" << time_rc99AECAL2 << " "
	 << "muon="  << time_rc99AMUON  << " "
	 << "sum="   << sum             << endl;
  }
  
  // mark
  ARTE_mark( "rc99AStandAlone 1", 0 );
  
  return 0;  
}
