
// *****************************************************************************
//
// source:    rcClassify.C
//
// type:      source code
//
// created:   14 Jun 2000
//	
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// symbol:    rcClassify
//
// called by: rcevnt_
//
// purpose:   event classification
//
// hints:     flagdst has 32 bits which are defined in eclass.hh.
//            Classification bits should be set by bitwise ORing (!)
//            flagdst with the appropriate mask (ECLASS_...) from eclass.hh.
//            Classification bits must NOT be set in other places besides
//            l4trig/l4tEvnt.C sets classification bits for randomly
//            (ECLASS_RNDTRG) and FLT (ECLASS_FLTRG) triggered events.
//
//            Event selection to output is done in l4trig/l4tEvnt.C.
//
// changes:
// 14 Jun 2000 (AG) creation
// 16 Jun 2000 (AG) add timing
// 20 Jun 2000 (AG) introduce special functions for various classes
//  6 Jul 2000 (AG) polished printout
// 19 Jul 2000 (AG) introduce VZERO classification
// 20 Jul 2000 (AG) introduce SLTEE, SLTMM, SLTDV classification
// 25 Jul 2000 (AG) added statistics
// 26 Jul 2000 (AG) use VZERO and VZERO2
// 27 Jul 2000 (AG) introduced SLTEEB
// 01 Aug 2000 (AG) polished code
// 02 Aug 2000 (AG) added rcClassDVTX, rcClassSLTHPT
// 03 Aug 2000 (AG) SLTHPT -> SLTCUT
// 09 Aug 2000 (AG) reset classification mask at the beginning
// 16 Aug 2000 (AG) introduced SLTYPS
// 06 Aug 2000 (AG) added VZERO2 0x10 -0x40
// 10 Nov 2000 (AG) introduced RECYPS and RECDV
// 14 Nov 2000 (AG) removed typo
// 15 Nov 2000 (AG) build classSum here
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/ArteUtil.h"
#include "farm/farm.hh"
#include "reco/eclass.hh"

// globals
int *classSum = 0;

// externals
extern int FADEB;                // debug flag
extern int rcSelectTracksStatus;

extern int classify;    // general
extern int classJPSIEE; // bit  1
extern int classJPSIMM; // bit  2
extern int classFCNC;   // bit  3
extern int classCAL1HT; // bit  4
extern int classVZERO ; // bit  5
extern int classVZERO2; // bit  6
extern int classDVTX;   // bit  7
extern int classSLTYPS; // bit  8
extern int classSLTCUT; // bit  9
extern int classSLTEEB; // bit 10
extern int classSLTEE;  // bit 11
extern int classSLTMM;  // bit 12
extern int classSLTDV;  // bit 13
extern int classTLT;    // bit 14
extern int classFLTTRG; // bit 15
extern int classRNDTRG; // bit 16
extern int classRECYPS; // bit 17
extern int classRECDV ; // bit 18
extern int classCLASS19; // bit 19
extern int classCLASS20; // bit 20
extern int classCLASS21; // bit 21
extern int classCLASS22; // bit 22
extern int classCLASS23; // bit 23
extern int classCLASS24; // bit 24
extern int classCLASS25; // bit 25
extern int classCLASS26; // bit 26
extern int classCLASS27; // bit 27
extern int classCLASS28; // bit 28
extern int classCLASS29; // bit 29
extern int classCLASS30; // bit 30
extern int classCLASS31; // bit 31
extern int classCLASS32; // bit 32

//
// function code
//
int
rcClassify()
{
  // show up
  ARTE_mark( "rcClassify 0", 0 );
  
  // EVHD table
  ArtePointer<EVHD> evhd = *( ArteTable<EVHD>::end() - 1 );
  
  // print flag
  if( FADEB ){
    cout << " CLASS: old flagdst=0x" << hex << evhd->flagdst << dec << endl;
  }

  // inits
  int rc = 0;
  static bool first = 1;
  if( first ){
    first = 0;

    // print
    cout << "CLASS: event classification bits" << endl;
    for( int bit=1; bit<=32; bit++ ){
      cout << "       " << bit   << " " << rcClassName( bit ) << endl;
    }
    
    // build array
    classSum = new int[32];
    for( int i=0; i<32; i++ ) classSum[i] = 0;
  }

  // check
  if( !classSum ) ARTE_exit( GRR_SFATAL, "rcClassify: No array, exit!", 0 );
  
  //===========================================================================
  // rest classification mask
  //===========================================================================
  evhd->flagdst = 0;

  //===========================================================================
  //  1 J/Psi -> e+e- ECLASS_JPSIEE
  //===========================================================================
  if( classJPSIEE ){
    ARTE_mark( "rcClassJPSIEE", 0 );
    rc = rcClassJPSIEE();
    if( rc ){
      classSum[1-1]++;
      evhd->flagdst |= ECLASS_JPSIEE; // bit-wise OR	
      cout << " CLASS: ECLASS_JPSIEE new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }

  //===========================================================================
  //  2 J/Psi -> mu+mu- ECLASS_JPSIMM
  //===========================================================================
  if( classJPSIMM ){
    ARTE_mark( "rcClassJPSIMM", 0 );
    rc = rcClassJPSIMM();
    if( rc ){
      classSum[2-1]++;
      evhd->flagdst |= ECLASS_JPSIMM; // bit-wise OR	
      cout << " CLASS: ECLASS_JPSIMM new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }
  
  //===========================================================================
  //  3 one high E_T calorimeter cluster ECLASS_CAL1HT
  //===========================================================================
  if( classCAL1HT ){
    ARTE_mark( "rcClassCAL1HT", 0 );
    rc = rcClassCAL1HT();
    if( rc ){
      classSum[3-1]++;
      evhd->flagdst |= ECLASS_CAL1HT; // bit-wise OR	
      cout << " CLASS: ECLASS_CAL1HT new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }

  //===========================================================================
  //  4 FCNC event ECLASS_FCNC
  //===========================================================================
  if( classFCNC ){
    ARTE_mark( "rcClassFCNC", 0 );
    rc = rcClassFCNC();
    if( rc ){
      classSum[4-1]++;
      evhd->flagdst |= ECLASS_FCNC; // bit-wise OR	
      cout << " CLASS: ECLASS_FCNC new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }

  //===========================================================================
  //  5 VZERO in event
  //===========================================================================
  if( classVZERO ){
    ARTE_mark( "rcClassVZERO", 0 );
    rc = rcClassVZERO();
    if( rc ){
      classSum[5-1]++;
      evhd->flagdst |= ECLASS_VZERO; // bit-wise OR	
      cout << " CLASS: ECLASS_VZERO new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }

  //===========================================================================
  //  6/7 VZERO2/DVTX in event
  //===========================================================================
  if( classVZERO2 || classDVTX ){
    ARTE_mark( "rcClassVZERO2", 0 );
    rc = rcClassVZERO2();

    // VZERO2 (Ks, K*, Phi)
    if( ( rc & 0x1 ) ||
	( rc & 0x2 ) ||
	( rc & 0x4 )    ){
      classSum[6-1]++;
      evhd->flagdst |= ECLASS_VZERO2; // bit-wise OR	
      cout << " CLASS: ECLASS_VZERO2 new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }

    // DVTX (D0, Ds, B, Bs)
    if( ( rc & 0x8  ) ||
	( rc & 0x10 ) ||
	( rc & 0x20 ) ||
	( rc & 0x40 ) || 
	( rc & 0x80 )    ){
      classSum[7-1]++;
      evhd->flagdst |= ECLASS_DVTX; // bit-wise OR	
      cout << " CLASS: ECLASS_DVTX new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }

  //===========================================================================
  //  8/9/10/11/12/13 SLT classification
  //===========================================================================
  if( classSLTYPS || classSLTCUT || classSLTEEB || classSLTEE || classSLTMM || classSLTDV ){
    ARTE_mark( "rcClassSLT", 0 );
    rc = rcClassSLT();

    // ll > 8.0 GeV 
    if( rc & 0x20 ){
      classSum[8-1]++;
      evhd->flagdst |= ECLASS_SLTYPS; // bit-wise OR
      cout << " CLASS: ECLASS_SLTYPS new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }

    // ll > 2.0 GeV 
    if( rc & 0x10 ){
      classSum[9-1]++;
      evhd->flagdst |= ECLASS_SLTCUT; // bit-wise OR
      cout << " CLASS: ECLASS_SLTCUT new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
    
    // ee brems
    if( rc & 0x8 ){
      classSum[10-1]++;
      evhd->flagdst |= ECLASS_SLTEEB; // bit-wise OR
      cout << " CLASS: ECLASS_SLTEEB new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
    
    // ee
    if( rc & 0x1 ){
      classSum[11-1]++;
      evhd->flagdst |= ECLASS_SLTEE; // bit-wise OR
      cout << " CLASS: ECLASS_SLTEE new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
    
    // mm  
    if( rc & 0x2 ){
      classSum[12-1]++;
      evhd->flagdst |= ECLASS_SLTMM; // bit-wise OR
      cout << " CLASS: ECLASS_SLTMM new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
    
    // dv
    if( rc & 0x4 ){
      classSum[13-1]++;
      evhd->flagdst |= ECLASS_SLTDV; // bit-wise OR
      cout << " CLASS: ECLASS_SLTDV new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }
  
  //===========================================================================
  // 14 TLT classification
  //===========================================================================
  if( classTLT ){
    ARTE_mark( "rcClassTLT", 0 );
    rc = rcClassTLT();
    if( rc ){
      classSum[14-1]++;
      evhd->flagdst |= ECLASS_TLT; // bit-wise OR
      cout << " CLASS: ECLASS_TLT new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }

  //===========================================================================
  // 15 FLT triggers
  //===========================================================================
  int tcode = faDDAQFCS_tcode();
  if( classFLTTRG ){
    if( ( tcode == 0 ) || ( tcode == 8 ) ){
      classSum[15-1]++;
      evhd->flagdst |= ECLASS_FLTTRG; // bit-wise OR
    }
  }
  
  //===========================================================================
  // 16 random triggers
  //===========================================================================
  if( classRNDTRG ){
    if( ( tcode == 8 ) || ( tcode == 9 ) ){
      classSum[16-1]++;
      evhd->flagdst |= ECLASS_RNDTRG; // bit-wise OR
    }
  }

  //===========================================================================
  // 17 high di-lepton mass classification
  //===========================================================================
  if( classRECYPS ){
    ARTE_mark( "rcClassRECYPS", 0 );
    rc = rcClassRECYPS();
    if( rc ){
      classSum[17-1]++;
      evhd->flagdst |= ECLASS_RECYPS; // bit-wise OR
      cout << " CLASS: ECLASS_RECYPS new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }


  //===========================================================================
  // 18 detached vertex classification
  //===========================================================================
  if( classRECDV ){
    ARTE_mark( "rcClassRECDV", 0 );
    rc = rcClassRECDV();
    if( rc ){
      classSum[18-1]++;
      evhd->flagdst |= ECLASS_RECDV; // bit-wise OR
      cout << " CLASS: ECLASS_RECDV new flagdst=0x" << hex << evhd->flagdst << dec << endl;
    }
  }

  // print
  if( FADEB ){
    cout << " CLASS: flagdst=0x" << hex << evhd->flagdst << dec << endl;
    cout << " CLASS: statistics:" << endl;
    cout << "  ";
    for( int i=0; i<15; i++ ){
      cout << "#" << i+1 << "=" << classSum[i] << " "; 
    }
    cout << endl;
    cout << "  ";
    for( int i=16; i<31; i++ ){
      cout << "#" << i+1 << "=" << classSum[i] << " "; 
    }
    cout << endl;
  }

  // show up
  ARTE_mark( "rcClassify 1", 0 );  

  // return
  return 0;    
}
